# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for performing hac cluster related operations on the commcell

HACClusters and HACCluster are 2 classes defined in this file

HACClusters:   Class for representing all the hac clusters associated with the commcell

HACCluster:    Class for a instance of a single hac cluster of the commcell


HACClusters
============

    __init__()                          --  initialize object of HAC clusters class associated with
    the commcell

    __str()                             --  returns all the HAC clusters of the commcell

    __repr__()                          --  returns the string to represent the instance

    __get_item()                        --  returns the details of HAC cluster for given cloud name

    _get_all_clusters()                 --  gets detail of all hac cluster associated to commcell

    _response_not_success()             --  raise exception when response is not 200

    get()                               --  return an HACCluster object for given cluster name

    has_cluster()                       --  returns whether the hac cluster is present or not in
    the commcell

    add()                            --  creates a new hac cluster to the commcell

    delete()                            --  deletes the hac cluster associated to commcell

    refresh()                           --  refresh the hac clusters details associated with commcell

HACClusters Attributes
-----------------------

    **all_hac_clusters**                 --  returns the dictionary consisting of all the hac clusters
    associated with the commcell and there details

HACCluster
============

    __init__()                          --  initialize object of IndexPool class

    __repr__()                          --  returns the string to represent the instance

    _response_not_success()             --  raise exception when response is not 200

    modify_node()                       --  methods to modify the HAC cluster node properties

    node_info()                         --  returns a dict consisting details of node present in the cluster

    refresh()                           --  refresh the index pool details associated with commcell

HACCluster Attributes
-----------------------

    **cluster_id**                      --  returns the cluster id for HAC cluster

    **cluster_name**                    --  returns the HAC cluster name

    **cloud_id**                        --  returns HAC cluster cloud id

    **node_names**                      --  returns a list of names of all HAC cluster nodes

"""

from copy import deepcopy
from past.builtins import basestring
from .exception import SDKException
from .datacube.constants import IndexServerConstants


class HACClusters(object):
    """Class for representing all the HAC clusters associated with the commcell."""

    def __init__(self, commcell_object):
        """Initialize object of the HACClusters class

            Args:
                commcell_object (object)    --  instance of class Commcell

            Returns:
                object  -   instance of class HACClusters
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._hac_group = None
        self._all_hac_clusters = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all HAC Clusters of the commcell.

                Returns:
                    str - string of all the HAC clusters associated with the commcell
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'HAC Name')
        index = 1
        for hac_name in self.all_hac_clusters:
            representation_string += '{:^5}\t{:^20}\n'.format(
                index, hac_name)
            index += 1
        return representation_string

    def __repr__(self):
        """Representation string for the instance of the HACClusters class."""
        return "HACClusters class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def __getitem__(self, value):
        """Returns the details of HAC cluster for given HAC name

            Args:
                value   (str)       --  name of HAC cluster

            Returns:
                dict    -   details of the HAC cluster

            Raises:
                HAC cluster not found
        """
        value = value.lower()
        if value.lower() in self.all_hac_clusters:
            return {"name": value.lower, "id": self.all_hac_clusters[value]}
        raise SDKException('HACClusters', '102')

    def _response_not_success(self, response):
        """Helper method to raise exception when response is not 200 (ok)

            Args:
                response    (object)    -   response object

            Raises:
                SDKException:
                    Response was not success
        """
        raise SDKException(
            'Response',
            '101',
            self._update_response_(
                response.text))

    def _get_all_clusters(self):
        """Gets details of all HAC clusters associated to the commcell"""
        if self._commcell_object.client_groups.has_clientgroup("HAC Cluster"):
            if self._hac_group is None:
                self._hac_group = self._commcell_object.client_groups.get("HAC Cluster")
            self._hac_group.refresh()
            for client_name in self._hac_group.associated_clients:
                client_obj = HACCluster(
                    self._commcell_object, client_name
                )
                self._all_hac_clusters[client_name.lower()] = int(client_obj.cloud_id)

    def has_cluster(self, hac_name):
        """Returns whether the HAC cluster with given name is present or not

            Args:
                hac_name    (str)       --  hac cluster name

            Returns:
                boolean     -   True if hac cluster is associated with the commcell
                else returns False

            Raises:
                SDKException:
                    Data type of the input(s) is not valid
        """
        if not isinstance(hac_name, basestring):
            raise SDKException('HACClusters', '101')
        return hac_name.lower() in self._all_hac_clusters

    def get(self, hac_name):
        """Returns instance of HACCluster class is cluster is found

            Args:
                hac_name        (str/int)   --      hac cluster name or id

            Returns:
                object          (HACCluster)   --  Instance of a single hac cluster

            Raises:
                SDKException:
                    Data type of the input(s) is not valid

                    HAC Cluster not found
        """
        if isinstance(hac_name, basestring):
            if hac_name.lower() in self.all_hac_clusters:
                return HACCluster(self._commcell_object, hac_name.lower())
        elif isinstance(hac_name, int):
            for cluster_name in self.all_hac_clusters:
                if int(self._all_hac_clusters[cluster_name]) == int(hac_name):
                    return HACCluster(self._commcell_object, cluster_name)
        else:
            raise SDKException('HACClusters', '101')
        raise SDKException("HACClusters", "102")

    def refresh(self):
        """Refreshes properties for HACClusters class"""
        self._commcell_object.clients.refresh()
        self._commcell_object.client_groups.refresh()
        self._all_hac_clusters = {}
        self._hac_group = None
        self._get_all_clusters()

    def add(self, cloud_name, cloud_node_names):
        """Creates a new HAC cluster

            Args:
                cloud_name      (str)       --  hac cluster cloud name
                cloud_node_names    (list)  --  string array of node names to be added to cluster

            Raises:
                SDKException:
                    Data type of the input(s) is not valid.

                    Response was not success.

                    Response was empty.

            Returns:
                Object  -   Instance of class HACCluster
        """
        if not (isinstance(cloud_name, basestring) and isinstance(cloud_node_names, list)):
            raise SDKException('HACClusters', '101')
        cloud_node_names = sorted(cloud_node_names)
        node_meta_infos = {
            'zkDataPort': '8091',
            'zkElectionPort': '8097',
            'zkListenerPort': '8090',
            'zkServerId': None,
            'zkDataDir': None
        }
        req_json = deepcopy(IndexServerConstants.REQUEST_JSON)
        del req_json['solrCloudInfo']
        del req_json['cloudMetaInfos']
        req_json['type'] = 6
        req_json['cloudInfoEntity'] = {
            "_type_": 169,
            "cloudName": cloud_name,
            "cloudDisplayName": cloud_name
        }
        server_id = 1
        for node_name in cloud_node_names:
            node_obj = self._commcell_object.clients.get(node_name)
            node_data = {
                "opType": IndexServerConstants.OPERATION_ADD,
                "status": 0,
                "nodeClientEntity": {
                    "_type_": 3,
                    "hostName": node_obj.client_hostname,
                    "clientName": node_name,
                    "clientId": int(node_obj.client_id)
                },
                "nodeMetaInfos": []
            }
            node_meta_infos['zkServerId'] = str(server_id)
            node_meta_infos['zkDataDir'] = node_obj.install_directory + "\\iDataAgent\\JobResults\\ZKData"
            for node_info in node_meta_infos:
                node_data['nodeMetaInfos'].append({
                    'name': node_info,
                    'value': node_meta_infos[node_info]
                })
            req_json['cloudNodes'].append(node_data)
            server_id += 1
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_CREATE'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json():
                if response.json()['genericResp'] == {} and \
                        'cloudId' in response.json():
                    self.refresh()
                    return HACCluster(self._commcell_object, cloud_name)
                else:
                    o_str = 'Failed to create HAC Cluster. Error: "{0}"'.format(
                        response.json()['genericResp'])
                    raise SDKException('Response', '102', o_str)
            else:
                raise SDKException('Response', '102')
        else:
            self._response_not_success(response)

    def delete(self, cloud_name):
        """Deletes an existing HAC cluster

        Args:
            cloud_name  (str)   --  HAC cluster cloud name to be deleted

        Raises:
            SDKException:
                Data type of the input(s) is not valid.

                Response was not success.

                Response was empty.
        """
        if not isinstance(cloud_name, basestring):
            raise SDKException('HACCluster', '101')
        cloud_id = self.all_hac_clusters[cloud_name.lower()]
        req_json = IndexServerConstants.REQUEST_JSON.copy()
        req_json["opType"] = IndexServerConstants.OPERATION_DELETE
        req_json['cloudInfoEntity']['cloudId'] = cloud_id
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_DELETE'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json() \
                    and 'errorCode' not in response.json()['genericResp']:
                self.refresh()
                return
            if response.json() and 'genericResp' in response.json():
                raise SDKException(
                    'Response', '102', response.json()['genericResp'].get(
                        'errorMessage', ''))
            raise SDKException('Response', '102')
        self._response_not_success(response)

    @property
    def all_hac_clusters(self):
        """Returns the details of all HAC clusters associated with commcell"""
        return self._all_hac_clusters


class HACCluster(object):
    """Class to perform HAC cluster operations on a specific HAC cluster"""

    def __init__(self, commcell_object, cluster_name, cluster_id=None):
        """Initializes the HACCluster class object

            Args:
                commcell_object     (object)        --  Instance of commcell class
                cluster_name        (str)           --  HAC cluster cloud name
                cluster_id          (int)           --  HAC cluster cloud id
                    default: None

            Returns:
                object  -   instance of the HACCluster class
        """
        self.commcell = commcell_object
        self._cluster_name = cluster_name
        self._cluster_id = cluster_id
        self._cluster_properties = None
        self.cluster_client_obj = None
        self.cluster_nodes = None
        self.refresh()

    def __repr__(self):
        """Representation string for the instance of the HACCluster class."""
        return "HACCluster class instance for Commcell: '{0}'".format(
            self.commcell.commserv_name
        )

    def _response_not_success(self, response):
        """Helper method to raise exception when response is not 200 (ok)

                Raises:
                    SDKException:
                        Response was not success
        """
        raise SDKException(
            'Response',
            '101',
            self.commcell._update_response_(
                response.text))

    def refresh(self):
        """Refreshes properties of the HAC cluster"""
        self.commcell.clients.refresh()
        if not self.commcell.clients.has_client(self._cluster_name):
            raise SDKException('HACClusters', '102')
        self.cluster_client_obj = self.commcell.clients.get(self._cluster_name)
        self._cluster_id = self.cluster_client_obj.client_id
        self._cluster_properties = self.cluster_client_obj.\
            properties['pseudoClientInfo']['distributedClusterInstanceProperties']['clusterConfig']['cloudInfo']
        self.cluster_nodes = self._cluster_properties['cloudNodes']

    def modify_node(self, node_name, listener_port=None, data_port=None,
                    election_port=None, data_dir=None):
        """Methods to modify the hac cluster node properties

        Args:
            node_name       (str)       -   Client name for the node
            listener_port   (int/str)   -   zkListenerPort address to be updated
                default - None              Sample: '8090' or 8090
            data_port       (int/str)   -   zkDataPort address to be updated
                default - None              Sample: '8091' or 8091
            election_port   (int/str)   -   zkElectionPort address to be updated
                default - None              Sample: '8097' or 8097
            data_dir        (str)       -   zoo keeper data directory
                default - None

        Raises:
            SDKException:
                HAC zKeeper node not found

                Response was not success

        Returns:
            None

        """
        node_info = self.node_info(node_name)
        port_infos = {}
        node_meta_info = node_info['nodeMetaInfos']
        for meta_info in node_meta_info:
            port_infos[meta_info['name']] = meta_info['value']
        if listener_port:
            port_infos['ZKLISTENERPORT'] = str(listener_port)
        if election_port:
            port_infos['ZKELECTIONPORT'] = str(election_port)
        if data_port:
            port_infos['ZKDATAPORT'] = str(data_port)
        if data_dir:
            port_infos['ZKDATADIR'] = str(data_dir)
        req_json = {
            'cloudId': self.cloud_id,
            'nodes': [{
                'status': 1,
                'opType': IndexServerConstants.OPERATION_EDIT,
                'nodeClientEntity': {
                    'clientId': int(self.commcell.clients[node_name]['id']),
                    'hostName': self.commcell.clients[node_name]['hostname'],
                    'clientName': node_name
                },
                'nodeMetaInfos': [
                    {
                        "name": "zkListenerPort",
                        "value": port_infos['ZKLISTENERPORT']
                    },
                    {
                        "name": "zkDataPort",
                        "value": port_infos['ZKDATAPORT']
                    },
                    {
                        "name": "zkElectionPort",
                        "value": port_infos['ZKELECTIONPORT']
                    },
                    {
                        "name": "zkDataDir",
                        "value": port_infos['ZKDATADIR']
                    }]
            }]
        }
        flag, response = self.commcell._cvpysdk_object.make_request(
            'POST', self.commcell._services['CLOUD_NODE_UPDATE'],
            req_json
        )
        if flag:
            if response.json() is not None:
                if 'errorCode' not in response.json():
                    self.refresh()
                    return
        self._response_not_success(response)

    def node_info(self, node_name):
        """Returns the hac cluster node information

        Args:
            node_name       (str)   -   HAC cluster node name

        Returns:
            dict        -   dictionary containing details of the hac node

        Raises:
            SDKException:
                HAC zKeeper node not found

        """
        for node_info in self.cluster_nodes:
            if node_info['nodeClientEntity']['clientName'].lower() == node_name.lower():
                return node_info
        raise SDKException('HACCluster', '103')

    @property
    def cloud_id(self):
        """Returns HAC cluster cloud id"""
        return self._cluster_properties['cloudInfoEntity']['cloudId']

    @property
    def node_names(self):
        """Returns a list of HAC cluster node names"""
        result = []
        for node_info in self.cluster_nodes:
            result.append(node_info['nodeClientEntity']['clientName'])
        return result

    @property
    def cluster_id(self):
        """Returns the HAC cluster pseudo client id"""
        return self._cluster_id

    @property
    def cluster_name(self):
        """Returns the HAC cluster cloud name"""
        return self._cluster_name



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

# pylint: disable=R1705

"""Helper file for session operations.

This file is used to perform Authentication for the user on the Commcell.

    #.  Check if the web server and service is valid and running

    #.  Perform Login operation to the Commcell using the credentials provided by the user

    #.  Store the Authtoken received after Login REST API call to use for the entire session

    #.  Renew Authtoken if credentials were given by the user during Commcell object
        initialization, and the current token has expired

    #.  Logout the current user from the Commcell, and disconnect the API session

    #.  Common method to be used in the entire SDK to perform REST API call on the Web Server


CVPySDK:

    __init__(commcell_object)   --  initialise object of the CVPySDK class and bind to the commcell

    _is_valid_service()         --  checks if the service is valid and running or not

    _login()                    --  sign in the user to the commcell with the credentials provided

    _renew_login_token()        --  renews the Authtoken for the currently logged in user

    _logout()                   --  sign out the current logged in user from the commcell,
    and ends the session

    _request()                  --  executes the request on the server and return the Response

    who_am_i()                  --  Fetches the username of the user to whom authtoken is mapped

    make_request()              --  run the http request specified on the URL/WebService provided,
    and return the flag specifying success/fail, and response

"""

from __future__ import absolute_import
from __future__ import unicode_literals

from xml.parsers.expat import ExpatError

import requests
import xmltodict

try:
    # Python 2 import
    import httplib
except ImportError:
    # Python 3 import
    import http.client as httplib

from .exception import SDKException


class CVPySDK(object):
    """Helper class for login, and logout operations.

        Also contains common method for running all HTTP requests.
    """

    def __init__(self, commcell_object, certificate_path=None):
        """Initialize the CVPySDK object for running various operations.

            Args:
                commcell_object     (object)    --  instance of the Commcell class


                certificate_path        (str)   --  path of the CA_BUNDLE or directory with
                certificates of trusted CAs (including trusted self-signed certificates)

                    default: None

            Returns:
                object  -   instance of the CVPySDK class

        """
        self._commcell_object = commcell_object
        self._certificate_path = certificate_path

    def _is_valid_service(self):
        """Checks if the service url is a valid url or not.

            Returns:
                True    -   if the service url is valid

                False   -   if the service url is invalid

            Raises:
                requests Connection Error:
                    requests.exceptions.ConnectionError

                requests Timeout Error:
                    requests.exceptions.Timeout

        """
        try:
            response = self._request(
                method='GET',
                url=self._commcell_object._web_service,
                timeout=184
            )

            # Valid service if the status code is 200 and response is True
            return response.status_code == httplib.OK and response.ok
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as error:
            raise error

    def _login(self):
        """Posts a login request to the server.

            Returns:
                str     -   Authtoken received from the WebServer upon successfull login

            Raises:
                SDKException:
                    if login failed

                    if response is empty

                    if response is not success

                requests Connection Error:
                    requests.exceptions.ConnectionError

        """
        try:
            if isinstance(self._commcell_object._password, dict):
                raise SDKException('CVPySDK', '104')


            if self._commcell_object.is_service_commcell:
                json_login_request = {
                    "mode": 4,
                    "clientType": 30,
                    "autoLogin": {
                        "autoLoginType": 5,
                        "encryptedMessage": self._commcell_object.master_saml_token
                    }
                }

            else:
                json_login_request = {
                    "mode": 4,
                    "username": self._commcell_object._user,
                    "password": self._commcell_object._password,
                    "deviceId": self._commcell_object.device_id,
                    "clientType": 30,
                }

            flag, response = self.make_request(
                'POST', self._commcell_object._services['LOGIN'], json_login_request
            )

            if flag:
                if response.json():
                    if "userName" in response.json() and "token" in response.json():
                        return response.json()['token']
                    else:
                        error_message = response.json()['errList'][0]['errLogMessage']
                        err_msg = 'Error: "{0}"'.format(error_message)
                        raise SDKException('CVPySDK', '101', err_msg)
                else:
                    raise SDKException('Response', '102')
            else:
                response_string = self._commcell_object._update_response_(response.text)
                raise SDKException('Response', '101', response_string)
        except requests.exceptions.ConnectionError as con_err:
            raise con_err

    def _renew_login_token(self):
        """Posts a Renew Login-Token request to the server.

            Returns:
                str     -   new token received from the WebServer

            Raises:
                SDKException:
                    if token renew failed

                    if response is empty

                    if response is not success

                requests Connection Error:
                    requests.exceptions.ConnectionError

        """
        try:
            if self._commcell_object._is_saml_login and not self._commcell_object.is_service_commcell:
                raise SDKException('CVPySDK', '106')

            token_renew_request = {
                "sessionId": self._commcell_object._headers['Authtoken'],
                "deviceId": self._commcell_object.device_id
            }

            flag, response = self.make_request(
                'POST', self._commcell_object._services['RENEW_LOGIN_TOKEN'], token_renew_request
            )

            if flag:
                if response.json():
                    if "token" in response.json():
                        return response.json()['token']
                    else:
                        error_message = response.json()['error']['errLogMessage']
                        err_msg = 'Error: "{0}"'.format(error_message)
                        raise SDKException('CVPySDK', '101', err_msg)
                else:
                    raise SDKException('Response', '102')
            else:
                response_string = self._commcell_object._update_response_(response.text)
                raise SDKException('Response', '101', response_string)
        except requests.exceptions.ConnectionError as con_err:
            raise con_err

    def _logout(self):
        """Posts a logout request to the server.

            Returns:
                str     -   response string from server upon logout success

        """
        flag, response = self.make_request('POST', self._commcell_object._services['LOGOUT'])

        if flag:
            self._commcell_object._headers['Authtoken'] = None

            if response.status_code == httplib.OK:
                return response.text
            else:
                return 'Failed to logout the user'
        else:
            return 'User already logged out'

    def _request(self, **kwargs):
        """Executes the request on the Server with the given parameters.

            If the certificate path is given and the Web Service starts with **https**,
            it adds the **verify** parameter to the request, and passes the certificate path as
            its value.

            Args:
                **kwargs    --  dict of keyword arguments, same as accepted by the

                    **requests.request** method

            Returns:
                object  -   **requests.Response** class instance, as received from calling the
                **requests.request** method

        """
        if self._certificate_path and self._commcell_object._web_service.startswith('https'):
            return requests.request(verify=self._certificate_path, **kwargs)

        return requests.request(**kwargs)

    def who_am_i(self, authtoken=None):
        """Get the username of the user, to whom the Authtoken belongs to.

            Args:
                authtoken   (str)   --  QSDK or SAML authentication token

            Returns:
                str     -   username of the user respective to the token

            Raises:
                SDKException:
                    if no user mapping found

        """
        temp_headers = self._commcell_object._headers.copy()

        if authtoken:
            temp_headers['Authtoken'] = authtoken

        flag, response = self.make_request(
            'POST', self._commcell_object._services['WHO_AM_I'], headers=temp_headers
        )

        if flag:
            if 'application/json' in response.headers['Content-Type']:
                if response.json().get('errorCode', 0) == 0:
                    return response.json()['user']['userName']
                else:
                    raise SDKException('CVPySDK', '107')
            else:
                user_dict = xmltodict.parse(response.content)

                if 'CvEntities_ProcessingInstructionInfo' in user_dict:
                    return user_dict['CvEntities_ProcessingInstructionInfo']['user']['@userName']
                else:
                    raise SDKException('CVPySDK', '107')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def make_request(
            self,
            method,
            url,
            payload=None,
            attempts=0,
            headers=None,
            stream=False,
            files=None):
        """Makes the request of the type specified in the argument 'method'.

            Args:
                method      (str)           --  HTTP operation to perform

                    e.g.:

                    -   GET

                    -   POST

                    -   PUT

                    -   DELETE

                url         (str)           --  the web url or service to run the HTTP request on


                payload     (dict / str)    --  data to be passed along with the request

                    default: None


                attempts    (int)           --  number of attempts made with the same request

                    default: 0


                headers     (dict)          --  dict of request headers for the request

                        if not specified we use default headers

                    default: None


                stream      (bool)          --  boolean specifying whether the request should get
                data via stream or normal get

                    default: False


                files       (dict)          --  file to upload in the form of

                        {
                            'file': open('report.txt', 'rb')
                        }

                    default: None

            Returns:
                tuple:
                    (True, response)    -   in case of success

                    (False, response)   -   in case of failure

            Raises:
                SDKException:
                    if the method passed is incorrect / not supported

                    if the number of attempts exceed 3

                requests Connection Error:
                    requests.exceptions.ConnectionError

        """
        try:
            if headers is None:
                headers = self._commcell_object._headers.copy()

            if method == 'POST':
                if isinstance(payload, (dict, list)):
                    if files is not None:
                        response = self._request(method=method, url=url, files=files, data=payload)
                    else:
                        response = self._request(
                            method=method, url=url, headers=headers, json=payload, stream=stream
                        )
                else:
                    try:
                        # call encode on the payload in case the characters in the payload
                        # are not encoded, and to encode the string payload to bytes
                        payload = payload.encode()
                    except AttributeError:
                        # pass silently if payload is alredy encoded in bytes
                        pass

                    if 'Content-type' in headers and headers['Content-type'] not in [
                            'application/x-www-form-urlencoded']:
                        try:
                            if payload is not None:
                                xmltodict.parse(payload)
                            headers['Content-type'] = 'application/xml'
                        except ExpatError:
                            headers['Content-type'] = 'text/plain'

                    response = self._request(
                        method=method, url=url, headers=headers, data=payload, stream=stream
                    )
            elif method == 'GET':
                response = self._request(method=method, url=url, headers=headers, stream=stream)
            elif method == 'PUT':
                response = self._request(method=method, url=url, headers=headers, json=payload)
            elif method == 'DELETE':
                response = self._request(method=method, url=url, headers=headers)
            else:
                raise SDKException('CVPySDK', '102', 'HTTP method {} not supported'.format(method))

            if response.status_code == httplib.UNAUTHORIZED and headers['Authtoken'] is not None:
                if attempts < 3:
                    self._commcell_object._headers['Authtoken'] = self._renew_login_token()
                    return self.make_request(method, url, payload, attempts + 1)
                else:
                    # Raise max attempts exception, if attempts exceeds 3
                    raise SDKException('CVPySDK', '103')

            if (response.status_code == httplib.OK or response.status_code == httplib.CREATED) and response.ok:
                return (True, response)
            else:
                return (False, response)
        except requests.exceptions.ConnectionError as con_err:
            raise con_err



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Service URLs for REST API operations.

SERVICES_DICT:
    A python dictionary for holding all the API services endpoints.

|

get_services(web_service):
    updates the SERVICES_DICT with the WebConsole API URL


To add a new REST API End-point to the SDK, user needs to add a key to the SERVICES_DICT_TEMPLATE
dictionary, for their usage, and the value will be in the format:

    "{0}{{ENDPOINT}}"

{0} will be replaced by the webconsole URL

"""

from __future__ import absolute_import
from __future__ import unicode_literals


SERVICES_DICT_TEMPLATE = {
    'LOGIN': '{0}Login',
    'LOGOUT': '{0}Logout',
    'RENEW_LOGIN_TOKEN': '{0}RenewLoginToken',
    'COMMSERV': '{0}CommServ',
    'GET_SAML_TOKEN': '{0}Commcell/SamlToken?validityInMins=%s',
    'WHO_AM_I': '{0}WhoAmI',

    'TFA': '{0}Commcell/Properties/TwoFactorAuth',
    'TFA_ENABLE': '{0}Commcell/Properties/TwoFactorAuth/Action/Enable',
    'TFA_DISABLE': '{0}Commcell/Properties/TwoFactorAuth/Action/Disable',
    'PRIVACY_ENABLE': '{0}Commcell/Properties/Privacy/Action/Enable',
    'PRIVACY_DISABLE': '{0}Commcell/Properties/Privacy/Action/Disable',
    'ORG_TFA': '{0}Organization/%s/TwoFactorAuth',
    'ORG_TFA_ENABLE': '{0}Organization/%s/TwoFactorAuth/Action/Enable',
    'ORG_TFA_DISABLE': '{0}Organization/%s/TwoFactorAuth/Action/Disable',
    'TFA_STATUS_OF_USER': '{0}Security/TwoFactorAuth/Status?username=%s',

    'GET_ALL_CLIENTS': '{0}Client',
    'GET_VIRTUAL_CLIENTS': '{0}Client?PseudoClientType=VSPseudo',
    'CLIENTFORCEDELETE':'{0}Client/%s?forceDelete=1',
    'CLIENT': '{0}Client/%s',
    'FILTER_CLIENTS':'{0}Client?%s',
    'GET_ALL_CLIENTS_PLUS_HIDDEN': '{0}Client?hiddenclients=true',
    'GET_ALL_PSEUDO_CLIENTS': '{0}Client?PseudoClientType',
    'CHECK_READINESS': '{0}Client/%s/CheckReadiness?network=%s&resourceCapacity=%s'
                       '&NeedXmlResp=true&includeDisabledClients=%s&CSCCNetworkCheck=%s',

    'GET_ALL_AGENTS': '{0}Agent?clientId=%s',
    'AGENT': '{0}Agent',
    'GET_AGENT': '{0}Agent?clientId=%s&applicationId=%s&propertyLevel=30',

    'GET_ALL_BACKUPSETS': '{0}Backupset?clientId=%s&propertyLevel=10',
    'BACKUPSET': '{0}Backupset/%s',
    'ADD_BACKUPSET': '{0}Backupset',

    'GET_ALL_INSTANCES': '{0}Instance?clientId=%s',
    'INSTANCE': '{0}Instance/%s',
    'APPLICATION_INSTANCE': '{0}Application/%s',
    'APPLICATION': '{0}Application',

    'GET_ALL_SUBCLIENTS': '{0}Subclient?clientId=%s&applicationId=%s&propertyLevel=20',
    'ADD_SUBCLIENT': '{0}Subclient',
    'SUBCLIENT': '{0}Subclient/%s',
    'SUBCLIENT_BACKUP': '{0}Subclient/%s/action/backup?backupLevel=%s',
    'VM_BACKUP': '{0}v2/vsa/vm/%s/backup?backupLevel=%s',
    'PREVIEW': '{0}Subclient/Content/Preview',

    'GET_JOBS': '{0}Job?clientId=%s&jobFilter=%s',
    'JOB': '{0}Job/%s',
    'JOB_DETAILS': '{0}JobDetails',
    'JOB_TASK_DETAILS': '{0}Job/%s/TaskDetails',
    'SUSPEND_JOB': '{0}Job/%s/action/pause',
    'RESUME_JOB': '{0}Job/%s/action/resume',
    'KILL_JOB': '{0}Job/%s/action/kill',
    'RESUBMIT_JOB': '{0}Job/%s/action/Resubmit',
    'ALL_JOBS': '{0}Jobs',
    'JOB_EVENTS': '{0}Events?jobId=%s',
    'JOB_MANAGEMENT_SETTINGS': '{0}CommServ/JobManagementSetting',

    'ENABLE_SHARED_LAPTOP': '{0}Commcell/Properties/SharedLaptopUsage/Action/Enable',
    'DISABLE_SHARED_LAPTOP': '{0}Commcell/Properties/SharedLaptopUsage/Action/Disable',

    'GET_MEDIA_AGENTS': '{0}V2/MediaAgents',
    'LIBRARY': '{0}Library',
    'GET_LIBRARY_PROPERTIES': '{0}Library/%s',
    'DETECT_TAPE_LIBRARY': '{0}Library?Action=detect',
    'CONFIGURE_TAPE_LIBRARY': '{0}Library?Action=configureTape',
    
    'LOCK_MM_CONFIGURATION': '{0}LockMMConfiguration',

    'STORAGE_POLICY': '{0}StoragePolicy',
    'GET_STORAGE_POLICY': '{0}StoragePolicy/%s',
    'GET_STORAGE_POLICY_ADVANCED': '{0}v2/StoragePolicy/%s?propertyLevel=10',
    'CREATE_STORAGE_POLICY_COPY': '{0}StoragePolicy?Action=createCopy',
    'DELETE_STORAGE_POLICY_COPY': '{0}StoragePolicy?Action=deleteCopy',
    'SCHEDULE_POLICY': '{0}SchedulePolicy',
    'CREATE_UPDATE_SCHEDULE_POLICY': '{0}Task',
    'GET_SCHEDULE_POLICY': '{0}SchedulePolicy/%s',
    'MEDIA_AGENT': '{0}MediaAgent/%s',
    'CLOUD_MEDIA_AGENT': '{0}MediaAgent/%s/CloudVMPowerManagement',
    'STORAGE_POLICY_COPY': '{0}V2/StoragePolicy/%s/Copy/%s',
    'STORAGE_POLICY_INFRASTRUCTUREPOOL': '{0}/StoragePolicy/Infrastructurepool?planId=%s',
    'RECOVERY_ENABLERS': '{0}MediaAgent/RecoveryEnabler?osType=CLIENT_PLATFORM_OSTYPE_UNIX ',

    'GET_ALL_ALERTS': '{0}AlertRule',
    'ALERT': '{0}AlertRule/%s',
    'CREATE_BLR_PAIR': '{0}Replications/Groups',
    'DELETE_BLR_PAIR': '{0}Replications/Monitors/continuous/%s',
    'GET_BLR_PAIRS': '{0}Replications/Monitors/continuous',
    'GET_BLR_PAIR': '{0}Replications/Monitors/continuous?replicationPairId=%s',
    'GET_BLR_PAIR_STATS': '{0}Replications/Statistics/%s',
    'GRANULAR_BLR_POINTS': '{0}Replications/Monitors/continuous/VmScale?destProxyClientId=%s&subclientId=%s&vmUuid=%s',
    'BLR_BOOT_DETAILS': '{0}/Replications/Monitors/continuous/Boot?replicationPairId=%s&bootType=%s&latest=true',
    'BROWSE_MOUNT_POINTS': '{0}/Client/%s/Action/BrowseMountPoints',

    'GET_VM_BROWSE': '{0}/VMBrowse?inventoryPath=%%5CFOLDER%%3AApplications%%3AApplications&PseudoClientId=%s',
    'GET_VOLUME_BROWSE': '{0}/VMBrowse?inventoryPath=%%5CFOLDER%%3AVolumes%%3AVolumes%%5CFOLDER%%3A%s%%3A%s&PseudoClientId=%s',
    'GET_APP_BROWSE': '{0}/VMBrowse?inventoryPath=%%5CFOLDER%%3AApplications%%3AApplications%%5CFOLDER%%3A%s%%3A%s&PseudoClientId=%s',
    'GET_LABEL_BROWSE': '{0}/VMBrowse?inventoryPath=%%5CFOLDER%%3ALabels%%3ALabels%%5CFOLDER%%3A%s%%3A%s&PseudoClientId=%s',

    'MODIFY_ALERT': '{0}AlertRule/%s/Action/Modify',
    'GET_ALL_CONSOLE_ALERTS': '{0}Alert?pageNo=%s&pageCount=%s',
    'GET_CONSOLE_ALERT': '{0}Alert/%s',
    'ENABLE_ALERT_NOTIFICATION': '{0}AlertRule/%s/notificationType/%s/Action/Enable',
    'DISABLE_ALERT_NOTIFICATION': '{0}AlertRule/%s/notificationType/%s/Action/Disable',
    'ENABLE_ALERT': '{0}AlertRule/%s/Action/Enable',
    'DISABLE_ALERT': '{0}AlertRule/%s/Action/Disable',
    'EMAIL_SERVER': '{0}EmailServer',

    'INVENTORY_SCHEDULES': '{0}Schedules?seaDataSourceId=%s&appType=132',
    'CLIENT_SCHEDULES': '{0}Schedules?clientId=%s',
    'AGENT_SCHEDULES': '{0}Schedules?clientId=%s&apptypeId=%s',
    'BACKUPSET_SCHEDULES': '{0}Schedules?clientId=%s&apptypeId=%s&backupsetId=%s',
    'INSTANCE_SCHEDULES':  '{0}Schedules?clientId=%s&apptypeId=%s&instanceId=%s',
    'SUBCLIENT_SCHEDULES': ('{0}Schedules?clientId=%s&apptypeId=%s&'
                            'backupsetId=%s&subclientId=%s'),
    'WORKFLOW_SCHEDULES': '{0}Schedules?workflowId=%s',
    'REPORT_SCHEDULES': '{0}/ScheduleReports',
    'OPTYPE_SCHEDULES': '{0}/Schedules?operationType=%s',
    'COMMCELL_SCHEDULES': '{0}/Schedules',
    'SCHEDULE': '{0}Schedules/%s',
    'ENABLE_SCHEDULE': '{0}Schedules/task/Action/Enable',
    'DISABLE_SCHEDULE': '{0}Schedules/task/Action/Disable',

    'LIVE_SYNC': '{0}Task',
    'LIVE_SYNC_DETAILS': '{0}/Task/%s/Details',

    'CLIENTGROUPS': '{0}ClientGroup',
    'CLIENTGROUP': '{0}ClientGroup/%s',

    'USERGROUPS': '{0}UserGroup?includeSystemCreated=true',
    'USERGROUP': '{0}UserGroup/%s',
    'DELETE_USERGROUP': '{0}UserGroup/%s?newUserId=%s&newUserGroupId=%s',
    'COMPANY_USERGROUP': '{0}UserGroup?parentProvider/providerId=%s',

    'BROWSE': '{0}DoBrowse',
    'RESTORE': '{0}CreateTask',
    'SQL_CLONES': '{0}sql/clones',
    'SQL_DATABASES': '{0}sql/databases?databaseName=%s',
    'SQL_DATABASE_DETAILS': '{0}sql/instance/%s/database/%s',

    'GET_WORKFLOWS': '{0}Workflow',
    'DEPLOY_WORKFLOW': '{0}Workflow/%s/action/deploy',
    'EXECUTE_WORKFLOW_API': '{0}Workflow/%s/action/execute',
    'EXECUTE_WORKFLOW': '{0}wapi/%s',
    'GET_WORKFLOW': '{0}Workflow/%s',
    'GET_WORKFLOW_DEFINITION': '{0}Workflow/%s/definition',
    'GET_INTERACTIONS': '{0}WorkflowInteractions',
    'GET_INTERACTION': '{0}Workflow/Interaction/%s',

    'INSTANCE_BROWSE': '{0}Client/%s/%s/Instance/%s/Browse',
    'CLOUD_DATABASE_BROWSE': '{0}BrowseRDSBackups',

    'SQL_RESTORE_OPTIONS': '{0}SQL/RestoreOptions',

    'EXECUTE_QCOMMAND': '{0}Qcommand/qoperation execute',
    'EXECUTE_QSCRIPT': '{0}Qcommand/qoperation execscript %s',
    'QCOMMAND': '{0}QCommand',
    'EXEC_QCOMMAND': '{0}ExecuteQCommand',

    'SOFTWARESTORE_DOWNLOADITEM': '{0}DownloadFile',
    'SOFTWARESTORE_PKGINFO': '{0}SoftwareStore/getPackagePublishInfo?packageName=%s',
    'SOFTWARESTORE_GETPKGID': '{0}SoftwareStoreItem',

    'CREATE_TASK': '{0}CreateTask',
    'ADD_INSTANCE': '{0}Instance',
    'MASKING_POLICY': '{0}MaskingPolicy',


    'GET_ANALYTICS_ENGINES': '{0}dcube/getAnalyticsEngine',
    'GET_ALL_DATASOURCES': '{0}dcube/GetDataSources?summary=1',
    'GET_DATASOURCE': '{0}dcube/getDataSource/%s',
    'GET_ALL_HANDLERS': '{0}dcube/getAllHandlers?dsId=%s',
    'GET_HANDLER': '{0}dcube/getHandler/?dsId=%s&handlerId=%s',
    'GET_CRAWL_HISTORY': '{0}dcube/GetHistory/%s',
    'GET_HANDLERS': '{0}dcube/gethandler?datasourceId=%s',
    'CREATE_HANDLER': '{0}dcube/savehandler',
    'GET_DATASOURCE_SCHEMA': '{0}dcube/getDSSchema/%s',
    'UPDATE_DATASOURCE_SCHEMA': '{0}dcube/updateschema',
    'GET_JDBC_DRIVERS': '{0}dcube/GetJDBCDrivers/%s',
    'DELETE_DATASOURCE_CONTENTS': '{0}dcube/deletedata/%s?softdelete=true',
    'DELETE_DATASOURCE': '{0}dcube/deleteDataSource/%s',
    'CREATE_DATASOURCE': '{0}dcube/createDataSource',
    'DATACUBE_IMPORT_DATA': '{0}dcube/post/%s/%s',
    'START_JOB_DATASOURCE': '{0}dcube/startjob/%s',
    'GET_STATUS_DATASOURCE': '{0}dcube/GetStatus/%s',
    'EXECUTE_HANDLER': '{0}dcube/handler/%s/%s?%s',
    'DELETE_HANDLER': '{0}dcube/deletehandler/%s',
    'SHARE_HANDLER': '{0}dcube/share/handler',
    'SHARE_DATASOURCE': '{0}dcube/share/datasource',
    'GET_CONTENT_ANALYZER_CLOUD': '{0}getContentAnalyzerCloud',
    'ACTIVATE_ENTITIES': '{0}dcube/entity',
    'ACTIVATE_ENTITY': '{0}dcube/entity/%s',
    'ACTIVATE_ENTITY_CONTAINER': '{0}EntityExtractionRules?getDisabled=true',
    'GET_TAGS': '{0}EDiscovery/Tags',
    'ADD_CONTAINER': '{0}PerformContainerOperation',
    'DELETE_CONTAINER': '{0}Containers/Action/Delete',
    'CA_UPLOAD_FILE': '{0}ContentAnalyzer/%s/action/uploadFile',
    'GET_CLASSIFIERS': '{0}dcube/classifiers?getDisabled=True',
    'START_TRAINING': '{0}ContentAnalyzer/%s/%s/ml/action/train',
    'CANCEL_TRAINING': '{0}ContentAnalyzer/%s/%s/training/cancel',

    'EDISCOVERY_INVENTORIES': '{0}EDiscoveryClients/Inventories',
    'EDISCOVERY_INVENTORY': '{0}EDiscoveryClients/Inventories/%s',
    'EDISCOVERY_ASSETS': '{0}EDiscoveryClients/Inventories/%s/Assets',
    'EDISCOVERY_CRAWL': '{0}EDiscoveryClients/Clients/%s/Jobs?datasourceId=%s&type=%s&operation=%s',
    'EDISCOVERY_JOBS_HISTORY': '{0}EDiscoveryClients/Clients/%s/Jobs/History?type=%s&datasourceId=%s',
    'EDISCOVERY_JOB_STATUS': '{0}EDiscoveryClients/Clients/%s/Jobs/Status?type=%s&datasourceId=%s',
    'EDISCOVERY_GET_DEFAULT_HANDLER': '{0}dcube/getdefaulthandler/%s',
    'EDISCOVERY_V2_GET_CLIENTS': '{0}V2/EDiscoveryClients/Clients?datasourceType=%s&clientGroup=%s&limit=%s&offset=%s&sortBy=%s&sortDir=%s',
    'EDISCOVERY_V2_GET_CLIENT_DETAILS': '{0}V2/EDiscoveryClients/Clients/%s?includeDocCount=%s&limit=%s&offset=%s&sortBy=%s&sortDir=%s&datasourceType=%s',
    'EDISCOVERY_V2_GET_CLIENT_GROUP_DETAILS': '{0}V2/EDiscoveryClients/ClientGroups/%s?includeDocCount=%s',
    'EDISCOVERY_SECURITY_ASSOCIATION': '{0}EDiscoveryClients/Security?appType=%s',
    'EDISCOVERY_DATA_SOURCES': '{0}V2/EDiscoveryClients/Datasources?datasourceId=%s&type=%s',
    'EDISCOVERY_DATA_SOURCE_DELETE': '{0}EDiscoveryClients/Datasources?datasourceId=%s&clientId=%s',
    'EDISCOVERY_DYNAMIC_FEDERATED': '{0}dcube/dynamicfederated/%s/%s/default',
    'EDISCOVERY_EXPORT': '{0}dcube/export/%s',
    'EDISCOVERY_EXPORT_STATUS': '{0}dcube/export/%s/status?token=%s',
    'EDISCOVERY_CREATE_DATA_SOURCE': '{0}EDiscoveryClients/Datasources',
    'EDISCOVERY_REVIEW_ACTIONS_WITH_REQUEST': '{0}EDiscoveryClients/Datasources/Actions/Requests',
    'EDISCOVERY_REVIEW_ACTIONS': '{0}V2/EDiscoveryClients/Datasources/Actions',
    'EDISCOVERY_CLIENTS': '{0}EDiscoveryClients?eDiscoverySubtype=%s',
    'EDISCOVERY_CLIENT_DETAILS': '{0}EDiscoveryClients/%s',
    'EDISCOVERY_DATA_SOURCE_STATS': '{0}EDiscoveryClients/Datasources?id=%s&type=%s&start=0&count=1000',
    'EDISCOVERY_CREATE_CLIENT': '{0}EDiscoveryClients/Clients',
    'EDISCOVERY_DELETE_CLIENT': '{0}EDiscoveryClients/Clients/%s',
    'EDISCOVERY_REQUESTS': '{0}V4/RequestManager/Request',
    'EDISCOVERY_REQUEST_DETAILS': '{0}V4/RequestManager/Request/%s',
    'EDISCOVERY_REQUEST_CONFIGURE': '{0}V4/RequestManager/Request/%s/Configure',
    'EDISCOVERY_REQUEST_PROJECTS': '{0}V4/RequestManager/Request/%s/Projects',
    'EDISCOVERY_REQUEST_FEDERATED': '{0}dcube/federated/%s/%s',
    'EDISCOVERY_REQUEST_DOCUMENT_MARKER': '{0}EDiscoveryClients/Tasks/%s/Documents',
    'EDISCOVERY_CONFIGURE_TASK': '{0}EDiscoveryClient/ConfigureTask',
    'EDICOVERY_TASK_WORKFLOW': '{0}EDiscoveryClients/Tasks/%s/Workflows',

    'GLOBAL_FILTER': '{0}GlobalFilter',
    'RESTORE_OPTIONS': '{0}Restore/GetDestinationsToRestore?clientId=0&appId=%s&flag=8',

    'UPLOAD_FULL_FILE': '{0}Client/%s/file/action/upload?uploadType=fullFile',
    'UPLOAD_CHUNKED_FILE': '{0}Client/%s/file/action/upload?uploadType=chunkedFile',

    'PLANS': '{0}V2/Plan',
    'PLAN': '{0}V2/Plan/%s',
    'DELETE_PLAN': '{0}V2/Plan/%s?confirmDelete=True',
    'ADD_USERS_TO_PLAN': '{0}V2/Plan/%s/Users',
    'GET_PLAN_TEMPLATE': '{0}V2/Plan/template?type=%s&subType=%s',
    'ELIGIBLE_PLANS': '{0}V2/Plan/Eligible?%s',
    'ASSOCIATED_ENTITIES': '{0}V2/Plan/%s/AssociatedEntities',
    'GET_PLANS': '{0}V2/Plan?type=%s&subType=%s',

    'DOMAIN_CONTROLER': '{0}CommCell/DomainController',
    'DELETE_DOMAIN_CONTROLER': '{0}CommCell/DomainController/%s',
    'DOMAIN_PROPERTIES': '{0}CommCell/DomainController?domainId=%s',

    'DRBACKUP': '{0}/CommServ/DRBackup',
    'DISASTER_RECOVERY_PROPERTIES': '{0}/Commcell/DRBackup/Properties',
    'CVDRBACKUP_STATUS': '{0}/cvdrbackup/status?commcellid=%s',
    'CVDRBACKUP_INFO': '{0}/cvdrbackup/info',
    'CVDRBACKUP_DOWNLOAD': '{0}/cvdrbackup/download',
    'CVDRBACKUP_REQUEST': '{0}cvdrbackup/requests',

    'ORACLE_INSTANCE_BROWSE': '{0}Instance/DBBrowse/%s',

    'METRICS': '{0}CommServ/MetricsReporting',
    'GET_METRICS': '{0}CommServ/MetricsReporting?isPrivateCloud=%s',

    'INTERNET_PROXY': '{0}/Commcell/InternetOptions/Proxy',

    'PASSWORD_ENCRYPTION_CONFIG': '{0}/Commcell/PasswordEncryptionConfig',

    'VM_ALLOCATION_POLICY': '{0}VMAllocationPolicy',
    'ALL_VM_ALLOCATION_POLICY': '{0}VMAllocationPolicy?showResourceGroupPolicy=true&deep=false&hiddenpolicies=true',
    'GET_VM_ALLOCATION_POLICY': '{0}VMAllocationPolicy/%s',
    'PROTECTED_VMS': "{0}VM?propertyLevel=AllProperties&status=1&fromTime=%s&toTime=%s&Limit=%s",
    'CONTINUOUS_REPLICATION_MONITOR': "{0}Replications/Monitors/continuous",
    'USERS': '{0}User',
    'USER': '{0}User/%s?Level=50',
    'DELETE_USER': '{0}User/%s?newUserId=%s&newUserGroupId=%s',
    'OTP': '{0}User/%s/preferences/OTP',

    'ROLES': '{0}Role',
    'ROLE': '{0}Role/%s',

    'ALL_CREDENTIALS': '{0}/CommCell/Credentials?propertyLevel=30',
    'ONE_CREDENTIAL': '{0}/CommCell/Credentials/%s?propertyLevel=30',
    'CREDENTIAL':   '{0}/Commcell/Credentials',
    'DELETE_RECORD': '{0}/Commcell/Credentials/action/delete',

    'GET_SECURITY_ROLES': '{0}Security/Roles',
    'SECURITY_ASSOCIATION': '{0}Security',
    'ENTITY_SECURITY_ASSOCIATION': '{0}Security/%s/%s',
    'GET_DATASTORE_BROWSE': '{0}VSBrowse/%s/%s?requestType=%s',

    'GET_DC_DATA': '{0}getDownloadCenterLookupData',
    'DC_ENTITY': '{0}saveDownloadCenterLookupEntities',
    'DC_SUB_CATEGORY': '{0}saveDownloadCenterSubCategory',
    'SEARCH_PACKAGES': '{0}searchPackages?release=11',
    'DOWNLOAD_PACKAGE': '{0}DownloadFile',
    'DOWNLOAD_VIA_STREAM': '{0}Stream/getDownloadCenterFileStream',
    'UPLOAD_PACKAGE': '{0}saveDownloadCenterPackage',
    'DELETE_PACKAGE': '{0}deleteDownloadCenterPackage?packageId=%s',

    'ORGANIZATIONS': '{0}Organization',
    'ORGANIZATION': '{0}Organization/%s',
    'UPDATE_ORGANIZATION': '{0}Organization?organizationId=%s',
    'GENERATE_AUTH_CODE': '{0}Organization/%s/Authtoken',
    'ACTIVATE_ORGANIZATION': '{0}Organization/%s/action/activate',
    'DEACTIVATE_ORGANIZATION': '{0}Organization/%s/action/deactivate',
    'ORGANIZATION_ASSOCIATION': '{0}company/%s/company-association',
    'ENABLE_PRIVACY_COMPANY_DATA': '{0}V2/Organization/%s/Privacy/Action/Lock',
    'DISABLE_PRIVACY_COMPANY_DATA': '{0}V2/Organization/%s/Privacy/Action/Unlock',
    'ORGANIZATION_TAGS' : '{0}Tags',
    'GET_ORGANIZATION_TAGS' : '{0}Tags/PROVIDER_ENTITY/%s',
    'COMPANY_PASSKEY' : '{0}Company/%s/Passkey',
    'COMPANY_AUTH_RESTORE' : '{0}Company/%s/AuthRestore',
    'EDIT_COMPANY_DETAILS': '{0}v4/company/%s',
    'CHECK_ELIGIBILITY_MIGRATION' : '{0}Company/%s/migration-entities',
    'MIGRATE_CLIENTS' : '{0}Company/%s/company-association',

    'STORAGE_POOL': '{0}StoragePool',
    'GET_STORAGE_POOL': '{0}StoragePool/%s',
    'ADD_STORAGE_POOL': '{0}StoragePool?Action=create',
    'DELETE_STORAGE_POOL': '{0}StoragePool/%s',
    'EDIT_STORAGE_POOL': '{0}StoragePool?Action=edit',
    'REPLACE_DISK_STORAGE_POOL': '{0}StoragePool?action=diskOperation',

    'KEY_MANAGEMENT_SERVER_ADD_GET': '{0}CommCell/KeyManagementServers',
    'KEY_MANAGEMENT_SERVER_DELETE': '{0}CommCell/KeyManagementServers/%s',

    'GET_ALL_MONITORING_POLICIES': '{0}logmonitoring/monitoringpolicy',
    'GET_ALL_ANALYTICS_SERVERS': '{0}AnalyticsServers',
    'GET_ALL_TEMPLATES': '{0}logmonitoring/search/getListOfTemplate',
    'CREATE_DELETE_EDIT_OPERATIONS': '{0}logmonitoring/policy/operations',
    'GET_MONITORING_POLICY_PROP': '{0}logmonitoring/Policy/%s',
    'RUN_MONITORING_POLICY': '{0}logmonitoring/Policy/%s/Action/Run',

    'LICENSE': '{0}CommcellRegistrationInformation',

    'REPLICATION_GROUPS': '{0}ReplicationGroups',
    'DELETE_REPLICATION_GROUPS': '{0}ReplicationGroups/Action/Delete',
    'REPLICATION_GROUP_DETAILS': '{0}Vsa/ReplicationGroup/%s',

    'DR_GROUPS': '{0}DRGroups',
    'GET_DR_GROUP': '{0}DRGroups/%s',
    'DR_GROUP_MACHINES': '{0}DRGroups/ClientList?source=1&entityType=3&entityId=%s',
    'DR_GROUP_JOB_STATS': '{0}DRGroups/JobStats?jobId=%s&drGroupId=%s&replicationId=%s&clientId=0',
    'DR_JOB_STATS': '{0}DRGroups/JobStats?jobId=%s',

    'REVERSE_REPLICATION_TASK': '{0}Replications/Monitors/streaming/Operation',
    'REPLICATION_MONITOR': '{0}Replications/Monitors/streaming?subclientId=0',
    'RPSTORE': '{0}Replications/RPStore',

    'CREATE_PSEUDO_CLIENT': '{0}pseudoClient',
    'CREATE_NAS_CLIENT': '{0}NASClient',
    'GET_OFFICE_365_ENTITIES': '{0}Office365/entities',
    'GET_DYNAMICS_365_CLIENTS': '{0}Office365/entities?agentType=5',
    'CLOUD_DISCOVERY': '{0}Instance/%s/CloudDiscovery?clientId=%s&appType=%s',
    'SET_USER_POLICY_ASSOCIATION': '{0}Office365/CloudApps/SetUserPolicyAssociation',
    'USER_POLICY_ASSOCIATION': '{0}Office365/CloudApps/UserPolicyAssociation',
    'UPDATE_USER_POLICY_ASSOCIATION': '{0}Office365/CloudApps/UpdateUserPolicyAssociation',
    'OFFICE365_MOVE_JOB_RESULT_DIRECTORY': '{0}Office365/MoveJobResultsDirectory',
    'OFFICE365_PROCESS_INDEX_RETENTION_RULES': '{0}Office365/ProcessIdxRetentionRules',
    'ADD_EXCHANGE': '{0}pseudoClient',
    'CREATE_CONFIGURATION_POLICIES': '{0}ConfigurationPolicies',
    'GET_CONFIGURATION_POLICIES': '{0}ConfigurationPolicies?policyType=email',
    'GET_CONFIGURATION_POLICIES_FS': '{0}ConfigurationPolicies?policyType=filesytem',
    'GET_CONFIGURATION_POLICY': '{0}ConfigurationPolicies/%s',
    'DELETE_CONFIGURATION_POLICY': '{0}ConfigurationPolicies/%s',
    'EMAIL_DISCOVERY': '{0}Backupset/%s/mailboxDiscover?discoveryType=%s',
    'EMAIL_DISCOVERY_WITHOUT_REFRESH': '{0}Backupset/%s/mailboxDiscover?discoveryType=%s&refreshMailboxDb=false',
    'GET_EMAIL_POLICY_ASSOCIATIONS': '{0}Subclient/%s/EmailPolicyAssociation?discoveryType=%s',
    'SET_EMAIL_POLICY_ASSOCIATIONS': '{0}/Subclient/EmailPolicyAssociation',

    'CREATE_NUTANIX_CLIENT': '{0}Client/Nutanix',

    'GET_EVENTS': '{0}Events',
    'GET_EVENT': '{0}Events/%s',

    'GET_ACTIVITY_CONTROL': '{0}CommCell/ActivityControl',
    'SET_ACTIVITY_CONTROL': '{0}CommCell/ActivityControl/%s/Action/%s',
    'SET_COMMCELL_PROPERTIES': '{0}Commcell/properties',

    'OPERATION_WINDOW': '{0}OperationWindow',
    'LIST_OPERATION_WINDOW': '{0}/OperationWindow/OpWindowList?clientId=%s',

    'RELEASE_LICENSE': '{0}Client/License/Release',
    'RECONFIGURE_LICENSE': '{0}Client/License/Reconfigure',
    'LIST_LICENSES': '{0}Client/%s/License',
    'GET_CLOUDAPPS_USERS': '{0}Instance/%s/CloudDiscovery?clientId=%s&eDiscoverType=%s',
    'GET_CLOUDAPPS_ONEDRIVE_USERS': '{0}Instance/%s/CloudDiscovery?clientId=%s&eDiscoverType=%s&subclientId=%s',
    'ENABLE_CLIENT_PRIVACY': '{0}/V3/Client/%s/Lock',
    'DISABLE_CLIENT_PRIVACY': '{0}/V3/Client/%s/Unlock',

    'IDENTITY_APPS': '{0}ThirdParty/App',

    'GLOBAL_PARAM': '{0}/setGlobalParam',
    'GET_GLOBAL_PARAM': '{0}/CommServ/AddRemoveSoftware/CommServeSoftwareCache',

    'SNAP_OPERATIONS': '{0}/Snaps/Operations',
    'STORAGE_ARRAYS': '{0}/StorageArrays',

    'GET_NETWORK_SUMMARY' :'{0}/FirewallSummary/%s',
    'NETWORK_TOPOLOGIES': '{0}FirewallTopology',
    'NETWORK_TOPOLOGY': '{0}FirewallTopology/%s',
    'PUSH_TOPOLOGY': '{0}FirewallTopology/%s/Push',

    'ADVANCED_JOB_DETAIL_TYPE': '{0}Job/%s/AdvancedDetails?infoType=%s',

    'CERTIFICATES': '{0}CommServ/Certificates',

    'GET_DAG_MEMBER_SERVERS': '{0}Exchange/DAG/%s/PseudoClientInfo',  # only for Exchange DAG
    'GET_RECOVERY_POINTS': '{0}Exchange/DAG/%s/RecoveryPoints?instanceId=%s&backupSetId=%s&subClientId=%s&appId=%s',

    'CASEDEFINITION': '{0}EDiscoveryClients/CaseDefinitions',

    'REGISTRATION': '{0}/RegFrgnCell',
    'UNREGISTRATION': '{0}/UnRegisterCommCell',
    'GET_REGISTERED_ROUTER_COMMCELLS': '{0}/CommCell/registered?getOnlyServiceCommcells=true',
    'GET_USERSPACE_SERVICE': '{0}/ServiceCommcell/UserSpace',
    'POLL_USER_SERVICE': '{0}/ServiceCommcell/IsUserPresent?userName=%s',
    'POLL_MAIL_SERVICE': '{0}/ServiceCommcell/IsUserPresent?email=%s',
    'POLL_REQUEST_ROUTER': '{0}/CommcellRedirect/RedirectListforUser?user=%s&getDistinctSAMLAppType=true',
    'MULTI_COMMCELL_DROP_DOWN': '{0}/MultiCommcellsForUser',
    'SERVICE_COMMCELL_ASSOC': '{0}/Security/MultiCommcell',
    'SYNC_SERVICE_COMMCELL': '{0}/RouterCommcell/SyncUserSpace?commcellGUID=%s',

    'DASHBOARD_ENVIRONMENT_TILE': '{0}clients/count?type=fileserver,vm,laptop',
    'DASHBOARD_NEEDS_ATTENTION_TILE': '{0}CommServ/Anomaly/Entity/Count?anomalousEntityType=14',

    'GET_ALL_LIVE_SYNC_PAIRS': '{0}Replications/Monitors/streaming?subclientId=%s',
    'GET_ALL_LIVE_SYNC_VM_PAIRS': '{0}Replications/Monitors/streaming?subclientId=%s&taskId=%s',
    'GET_LIVE_SYNC_VM_PAIR': '{0}Replications/Monitors/streaming?subclientId=%s&replicationPairId=%s',

    'BACKUP_NETWORK_PAIRS': '{0}CommServ/DataInterfacePairs?ClientId=%s',
    'BACKUP_NETWORK_PAIR': '{0}CommServ/DataInterfacePairs',

    'GET_ALL_RECOVERY_TARGETS':
        '{0}/VMAllocationPolicy?showResourceGroupPolicy=true&showNonResourceGroupPolicy=false&deep=true',
    'GET_RECOVERY_TARGET': '{0}/VMAllocationPolicy/%s',

    'RETIRE': '{0}Client/%s/Retire',

    'DATASOURCE_ACTIONS': '{0}EDiscoveryClients/Datasources/Actions',
    'CLOUD_CREATE': '{0}cloud/create',
    'CLOUD_MODIFY': '{0}cloud/modify',
    'CLOUD_DELETE': '{0}cloud/delete',
    'CLOUD_ROLE_UPDATE': '{0}cloud/role/update',
    'CLOUD_NODE_UPDATE': '{0}cloud/node/update',
    'GET_ALL_INDEX_SERVERS': '{0}dcube/getAnalyticsEngine?retrieveall=true',
    'GET_ALL_ROLES': '{0}IndexingGateway/GetAnalyticsRolesInfo',
    'GET_SWAGGER': '{0}swagger/V3/swagger.json',

    'COMMCELL_METADATA': '{0}Commcell/MetaData',
    'METALLIC_LINKING': '{0}CloudService/Subscription',
    'CV_METALLIC_LINKING': '{0}/CloudService/Subscription/Details',
    'METALLIC_COMPLETED_SETUPS': '{0}CloudService/CompletedSetups',
    'USER_MAPPINGS': '{0}GetUserMappings',
    'METALLIC_REGISTERED': '{0}CloudServices/Registered',
    'METALLIC_UNLINK': '{0}CloudService/Unsubscribe'
}


def get_services(web_service):
    """Initializes the SERVICES DICT with the web service for APIs.

        Args:
            web_service     (str)   --  web service string for APIs

        Returns:
            dict    -   services dict consisting of all APIs

    """
    services_dict = SERVICES_DICT_TEMPLATE.copy()
    for service in services_dict:
        services_dict[service] = services_dict[service].format(web_service)

    return services_dict



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for commvault entity creation.

Entity:
=======
    __init__()      --  initialise object of Entity class

"""

import ntpath
import posixpath

from cvpysdk.commcell import Commcell

import SimpanaDefaults


class Entity(object):
    """ Class for browse operation """

    # Instance commvault pacakge is installed on
    package_instance = None

    # Commvault entity details
    commserv_name = None
    client_name = None
    agent_name = None
    backupset_name = None
    subclient_name = None

    # Commvault entity objects
    commcell = None
    client = None
    agent = None
    backupset = None
    subclient = None

    # Simpana defaults
    simpana_default_obj = SimpanaDefaults.Simpanadefaults()
    separator = None

    def __init__(
            self,
            logger_object,
            commcell_object,
            client_name,
            agent_name,
            backupset_name=None,
            subclient_name=None,
            package_instance=None
    ):
        """ Initialize the browse class instance

            Args:
                logger_object   (str)   -- Instance of Logger class

                commcell_object (obj)   -- instance of commcell class

                client_name     (str)   -- Client to perform operations on

                agent_name      (str)   -- Agent to perform operations on

                backupset_name  (str)   -- Backupset to be used

                subclient_name  (str)   -- Subclient to be used

                package_instance   (str)   -- Instance where commvault is installed

        """
        assert isinstance(commcell_object, Commcell) and isinstance(client_name, str) and isinstance(
            agent_name, str), "Commvault entity creation failed"

        try:
            Entity.commcell = commcell_object
            if Entity.commserv_name != commcell_object.commserv_name:
                Entity.commserv_name = commcell_object.commserv_name
                Entity.client_name = None
                Entity.agent_name = None
                Entity.backupset_name = None
                Entity.subclient_name = None

            if Entity.client_name != client_name:
                Entity.client = commcell_object.clients.get(client_name)
                logger_object.info("Client object created successfully")
                Entity.separator = '\\' if 'windows' in Entity.client.os_info.lower() else '/'
                Entity.client_name = client_name
                Entity.agent_name = None
                Entity.backupset_name = None
                Entity.subclient_name = None

            if Entity.agent_name != agent_name:
                Entity.agent = Entity.client.agents.get(agent_name)
                logger_object.info("Agent object created successfully")
                Entity.agent_name = agent_name
                Entity.backupset_name = None
                Entity.subclient_name = None

            if not backupset_name:
                backupset_name = Entity.agent.backupsets.default_backup_set

            if Entity.backupset_name != backupset_name:
                Entity.backupset = Entity.agent.backupsets.get(backupset_name)
                logger_object.info("Backupset object created successfully")
                Entity.backupset_name = backupset_name
                Entity.subclient_name = None

            if subclient_name and Entity.subclient_name != subclient_name:
                Entity.subclient = Entity.backupset.subclients.get(subclient_name)
                logger_object.info("subclient object created successfully")
                Entity.subclient_name = subclient_name

            if package_instance:
                Entity.package_instance = package_instance

            if not Entity.package_instance:
                Entity.package_instance = Entity.simpana_default_obj.get_local_instance_name(
                    Entity.client.install_directory)

        except Exception as exp:
            logger_object.info("Exception occurred in Entity object creation\nError: %s", exp)
            raise Exception("Exception occurred in Entity object creation")

    @staticmethod
    def _get_absolute_path(path, live_path, vsa_guid=''):
        """ Returns the absolute path of the file

        Args:
            path    (str)   -- Absolute/Relative path of the file

            live_path    (str)   -- Live path of the tool

            vsa_guid  (str)     --  Guid for VSA VM

        Returns:
            (str)   -- Absolute path of the file

        """
        if not path:
            return ''
        if Entity.client:
            if 'windows' in Entity.client.os_info.lower():
                if ntpath.isabs(path):
                    return path
            elif posixpath.isabs(path):
                if vsa_guid:
                    return '\\{0}{1}'.format(vsa_guid.rstrip("\n"), path.replace('/', '\\'))
                return path

        if live_path:
            try:
                stack = live_path.split(Entity.separator)

                for temp in path.split(Entity.separator):
                    if temp != '.':
                        if temp == '..':
                            stack.pop()
                        elif temp:
                            stack.append(temp)

                return '{0}'.format(Entity.separator).join(stack)

            except Exception:
                print('Failed to get absolute path of the file, please provide full path')
                raise Exception("Failed to get the absolute path of the file")
        else:
            print('Current working directory not set, please specify absolute path of file')
            raise Exception("Current working directory not set")



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing deduplication engine related operations on the commcell.

This file has all the classes related to deduplication engine operations.

DeduplicationEnigines:  Class for representing all the deduplication engines associated to the commcell.

DeduplicationEninge:    Class for representing a single deduplication engine associated to the commcell.

Store:  Class for representing a single deduplication store associated to the deduplication engine.

SubStore:   Class for representing a single substore associated to the deduplication store.

DeduplicationEngines:
    __init__(commcell_object)   - Initialise the DeduplicationEngines class instance


    __repr__()                  - Representation string consisting of deduplication engines.

    __str__()                   - Representation string for the instance of the DeduplicationEngines class.

    _get_engines()              - returns all engines properties on commcell

    get()                       - returns list of all engines

    has_engine()                - checkes if a engine exisits for storage policy and copy name

    refresh()                   - refreshes all engine properties

DeduplicationEngine:
    __init__(commcell_object, storage_policy_name, copy_name)   - Initialise the DeduplicationEngine class instance

    __repr__()                  - Representation string consisting of deduplication engine.

    __str__()                   - Representation string for the instance of the DeduplicationEngine class.

    _initialize_policy_and_copy_id()    - Gets deduplication engine properties

    _get_engine_properties()    - initializes deduplication engine properties

    _initialize_stores()        - initializes all the stores presnet in deduplication engine

    refresh()                   - refreshes all the deduplication engine properties

    all_stores()                - Checks if a deduplication store exists in a engine with provided storeid id.

    has_store()                 - returns list of all stores present in deduplication engines

    get()                       - Returns store class object for sssthe store id on deduplication engine

Store:
    __init__(commcell_object, storage_policy_name, copy_name, store_id)   - Initialise the Store class instance

    __repr__()                  - Representation string consisting of deduplication store.

    __str__()                   - Representation string for the instance of the Store class.

    _initialize_store_properties()  - initializes store properties

    _get_substores()            - gets all substores in a store along with properties

    _initialize_substores()     - initializes all substore properties

    refresh()                   - refreshes store properties

    all_substores()             - returns list of all substores in a store

    has_substore()              - checks if a substore exists in a store

    get()                      - gets a substore class object for provided substore id

    seal_deduplication_database() - Seals the deduplication database

    recover_deduplication_database()    - starts DDB Reconstruction job for store

    run_space_reclaimation()    - starts DDB space reclaimation job for store

    run_ddb_verification()      - starts DDB verification job for store

    enable_store_pruning        -  Property to get the current state of store pruning or enable/disable store pruning

Substore:
    __init__(commcell_object, storage_policy_name, copy_name,
            store_id, substore-id)   - Initialise the SubStore class instance

    __repr__()                  - Representation string consisting of substore.

    __str__()                   - Representation string for the instance of the SubStore class.

    _initialize_substore_properties()   - initialize substore properties of a store

    refresh()                   - refreshes substore properties

    mark_for_recovery()         - marks a substore for recovery
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from past.builtins import basestring
from future.standard_library import install_aliases
from enum import Enum

from .exception import SDKException
from .job import Job

install_aliases()


class StoreFlags(Enum):
    IDX_SIDBSTORE_FLAGS_PRUNING_ENABLED = 536870912


class DeduplicationEngines(object):
    """Class for getting all the deduplication engines associated with the commcell."""

    def __init__(self, commcell_object):
        """Initialize object of the DeduplicationEngines class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the StoragePolicies class
        """
        self._commcell_object = commcell_object

        self._engines = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all deduplication engines of the commcell.

            Returns:
                str - string of all the deduplication associated with the commcell
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'Deduplication engine')
        for index, engine in enumerate(self._engines):
            sub_str = '{:^5}\t{}/{}\n'.format(index + 1, engine[0], engine[1])
            representation_string += sub_str
        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the DeduplicationEngines class."""
        return "DeduplicationEngines class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def _get_engines(self):
        """
        gets all the deduplication engines associated with the commcell

        Return:
            dict - consists of all the engines in the commcell

        Raises:
            SDKException:
                if response is not success
        """
        request_json = {
            "EVGui_DDBEnginesReq": {
                "filterOptions": {
                    "propertyLevel": 1
                },
                "storagepolicy": {
                    "storagePolicyId": 0
                },
                "spCopy": {
                    "copyId": 0,
                    "storagePolicyId": 0
                },
                "store": {
                    "type": 115
                },
                "flags": 0
            }
        }
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response and response.json() and 'engines' in response.json():
                engines = response.json()['engines']

                if engines == []:
                    return {}

                engines_dict = {}

                for engine in engines:
                    temp_sp_name = engine['sp']['name'].lower()
                    temp_sp_id = str(engine['sp']['id']).lower()
                    temp_copy_name = engine['copy']['name'].lower()
                    temp_copy_id = str(engine['copy']['id']).lower()
                    engines_dict[(temp_sp_name, temp_copy_name)] = [temp_sp_id, temp_copy_id]
                return engines_dict
            return {}
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    @property
    def all_engines(self):
        """returns list all the deduplication engines with storage polic and copy name"""
        return list(self._engines.keys())

    def refresh(self):
        """refreshes all the deduplication engines and their properties"""
        self._engines = self._get_engines()

    def has_engine(self, storage_policy_name, copy_name):
        """Checks if a deduplication engine exists in the commcell with the input storage policy and copy name.

            Args:
                storage_policy_name (str)  --  name of the storage policy

                copy_name (str)  --  name of the storage policy copy

            Returns:
                bool - boolean output whether the deduplication engine exists in the commcell or not

            Raises:
                SDKException:
                    if type of the storage policy and copy name arguments are not string
        """
        if not isinstance(storage_policy_name, basestring) and not isinstance(copy_name, basestring):
            raise SDKException('Storage', '101')
        return self._engines and (storage_policy_name.lower(), copy_name.lower()) in self._engines

    def get(self, storage_policy_name, copy_name):
        """
        Returns eng class object for the engine on deduplication engines

        Args:
            storage_policy_name (str) - name of the storage policy

            copy_name (str) - name of the deduplication enabled copy

        Return:
             object - instance of engine class for a given storage policy and copy name

        Raises:
            SDKException:
                if type of arguments are not string

                if no engine exists with given storage policy and copy name
        """
        if not isinstance(storage_policy_name, basestring) and  not isinstance(copy_name, basestring):
            raise SDKException('Storage', '101')

        storage_policy_name = storage_policy_name.lower()
        copy_name = copy_name.lower()

        if self.has_engine(storage_policy_name, copy_name):
            return DeduplicationEngine(
                self._commcell_object, storage_policy_name, copy_name
            )
        raise SDKException(
            'Storage', '102', f'No dedupe engine exists with name: {storage_policy_name}/{copy_name}'
        )


class DeduplicationEngine(object):
    """Class to get all stores associated for deduplication engine"""

    def __init__(self, commcell_object, storage_policy_name, copy_name, storage_policy_id=None, copy_id=None):
        """Initialise the DeduplicationEngine class instance.

        Args:
            commcell_object (object)    - instance of class Commcell

            storage_policy_name (str)   - storage policy name on commcell

            copy_name (str)             - copy name under storage policy

            storage_policy_id (int)     - storage policy id for commcell

            copy_id (int)               - copy id under storage policy
        """
        self._storage_policy_name = storage_policy_name.lower()
        self._copy_name = copy_name.lower()
        self._commcell_object = commcell_object
        self._engine_properties = {}
        self._stores = {}
        if not storage_policy_id and not copy_id:
            self._initialize_policy_and_copy_id()
        else:
            if not isinstance(storage_policy_id, int) and not isinstance(copy_id, int):
                raise SDKException('Storage', '101')
            self._storage_policy_id = storage_policy_id
            self._copy_id = copy_id
        self.refresh()

    def __str__(self):
        """Representation string consisting of deduplication engine.

            Returns:
                str - string of all the stores associated with the deduplication engine
        """
        representation_string = '{:^5}\t{:^20}\t{}\n\n'.format('Store ID.', 'Store', 'Sealed Status')

        for store_id in self._stores:
            status = 'sealed' if self._stores[store_id]['sealedTime'] else 'active'
            sub_str = '{:^5}\t{:20}\t{}\n'.format(store_id, self._stores[store_id]['storeName'], status)
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the DeduplicationEngine class."""
        return "DeduplicationEngine class instance for Engine: '{0}/{1}'".format(
            self._storage_policy_name, self._copy_name
        )

    def _initialize_policy_and_copy_id(self):
        """initializes the variables for storage policy and copy id"""
        deduplication_engines = DeduplicationEngines(self._commcell_object)
        policy_and_copy_id = deduplication_engines._engines[(self._storage_policy_name, self._copy_name)]
        self._storage_policy_id = policy_and_copy_id[0]
        self._copy_id = policy_and_copy_id[1]

    def _get_engine_properties(self):
        """
        Gets deduplication engine properties

        Return:
             dict - engine properties for each store on deduplication engine

        Raises:
            SDKException:
                    if response is empty

                    if response is not success
        """
        request_json = {
            "EVGui_DDBEnginesReq": {
                "filterOptions": {
                    "propertyLevel": 20
                },
                "storagepolicy": {
                    "storagePolicyId": self.storage_policy_id
                },
                "spCopy": {
                    "copyId": self.copy_id,
                    "storagePolicyId": self.storage_policy_id
                },
                "store": {
                    "type": 115
                },
                "flags": 0
            }
        }

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response and response.json():
                return response.json()
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def _initialize_engine_properties(self):
        """initializes deduplication engine properties"""
        self._engine_properties = self._get_engine_properties()
        self._initialize_stores()

    def _initialize_stores(self):
        """initializes all the stores presnet in deduplication engine"""
        for store in self._engine_properties.get('engines'):
            temp_sp_name = store['sp']['name'].lower()
            temp_copy_name = store['copy']['name'].lower()
            if (temp_sp_name, temp_copy_name) == (self._storage_policy_name, self._copy_name):
                self._stores[store['storeId']] = store

    def refresh(self):
        """refreshes all the deduplication engine properties"""
        self._initialize_engine_properties()

    @property
    def all_stores(self):
        """returns list of all stores present in deduplication engines"""
        stores = []
        for store_id in self._stores:
            status = 'sealed' if self._stores[store_id]['sealedTime'] else 'active'
            stores.append([store_id, self._stores[store_id]['storeName'], status])
        return stores

    @property
    def storage_policy_id(self):
        """returns storage policy id associated to engine"""
        return self._storage_policy_id

    @property
    def copy_id(self):
        """returns copy id associated to engine"""
        return self._copy_id

    def has_store(self, store_id):
        """Checks if a deduplication store exists in a engine with provided storeid id.
        Args:
            store_id (int) - deduplication store id to check existance

        Returns:
            bool - boolean output whether the deduplication store exists in the engine or not

        Raises:
            SDKException:
                if type of the store id argument is not int
        """
        if not isinstance(store_id, int):
            raise SDKException('Storage', '101')
        return self._stores and store_id in self._stores

    def get(self, store_id):
        """
        Returns store class object for the store id on deduplication engine

        Args:
            store_id (int) - id of the store on deduplication engine

        Return:
            object - instance of Store class for a given store id

        Raises:
            if type of store id argument is not integer

            if no store exists with given store id
        """
        if not isinstance(store_id, int):
            raise SDKException('Storage', '101')

        if self.has_store(store_id):
            return Store(self._commcell_object, self._storage_policy_name, self._copy_name, store_id)
        raise SDKException(
            'Storage', '102', f'No store exists with id: {store_id}'
        )


class Store(object):
    """Class for performing deduplication store level operations for deduplication engine"""

    def __init__(self, commcell_object, storage_policy_name, copy_name, store_id):
        """Initialise the Store class instance.

        Args:
            commcell_object (object)    - commcell class instance

            storage_policy_name (str)   - storage policy name in commcell

            copy_name (str)             - copy name under storage policy

            store_id (int)              - deduplication store id in commcell
        """
        self._storage_policy_name = storage_policy_name.lower()
        self._copy_name = copy_name.lower()
        self._store_id = store_id
        self._commcell_object = commcell_object
        self._substores = {}
        self._store_properties = {}
        self._extended_flags = None
        self._dedupe_flags = None
        self._store_flags = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of deduplication store.

            Returns:
                str - string of all the substores associated with the deduplication store
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('SubStore ID.', 'SubStoreList')

        for substore_id in self._substores:
            sub_str = '{:^5}\t[{}]{}\n'.format(substore_id, self._substores[substore_id]['MediaAgent']['name'],
                                               self._substores[substore_id]['Path'])
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the Store class."""
        return "Store class instance for Deduplication Store ID: '{0}'".format(self._store_id)

    def _initialize_store_properties(self):
        """initializes the deduplication store proerties"""
        deduplication_engine = DeduplicationEngine(self._commcell_object, self._storage_policy_name, self._copy_name)
        self._store_properties = deduplication_engine._stores[self._store_id]
        self._extended_flags = self._store_properties['storeExtendedFlags']
        self._dedupe_flags = self._store_properties['dedupeFlags']
        self._store_flags = self._store_properties['storeFlags']
        self._initialize_substores()

    def _get_substores(self):
        """
        Gets properties of all the substores in a deduplication store

        Return:
             dict - store properties and substore list for each substore on deduplication store

        Raises:
            SDKException:
                    if response is empty

                    if response is not success
        """
        request_json = {
            "EVGui_SubStoreListReq": {
                "commcellId": 2,
                "storeId": self.store_id
            }
        }

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response.json():
                return response.json()
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def _initialize_substores(self):
        """initialisez all the substores present in a deduplication store"""
        substre_raw = self._get_substores()
        for substore in substre_raw.get('subStoreList'):
            self._substores[substore['subStoreId']] = substore

    def refresh(self):
        """refreshes all the deduplication store properties"""
        self._initialize_store_properties()

    @property
    def all_substores(self):
        """returns list of all substores present on a deduplication store"""
        substores = []
        for substore_id in self._substores:
            substores.append([substore_id, self._substores[substore_id]['Path'],
                              self._substores[substore_id]['MediaAgent']['name']])
        return substores

    def has_substore(self, substore_id):
        """Checks if a substore exists in a deduplication store with provided substore id.
        Args:
            substore_id (int) - substore id to check existance

        Returns:
            bool - boolean output whether the substore exists in the store or not

        Raises:
            SDKException:
                if type of the store id argument is not int
        """
        if not isinstance(substore_id, int):
            raise SDKException('Storage', '101')
        return self._substores and substore_id in self._substores

    def get(self, substore_id):
        """
        Returns substore class object for the substore id on deduplication store

        Args:
            substore_id (int) - id of the substore on deduplication store

        Return:
             object - instance of subStore class for a given substore id

        Raises:
            if type of substore id argument is not integer

            if no substore exists with given substore id
        """
        if not isinstance(substore_id, int):
            raise SDKException('Storage', '101')

        if self.has_substore(substore_id):
            return SubStore(self._commcell_object, self._storage_policy_name, self._copy_name, self._store_id,
                            substore_id)
        raise SDKException(
            'Storage', '102', f'No substore exists with id: {substore_id}'
        )

    @property
    def store_flags(self):
        """returns the deduplication flags on store"""
        self.refresh()
        return self._store_flags

    @property
    def store_name(self):
        """returns the store display name"""
        return self._store_properties.get('storeName')

    @property
    def store_id(self):
        """return the store id"""
        return self._store_id

    @property
    def version(self):
        """returns deduplication store version"""
        return self._store_properties.get('ddbVersion')

    @property
    def status(self):
        """returns the store display name"""
        return self._store_properties.get('status')

    @property
    def storage_policy_name(self):
        """returns storage policy name associated with store"""
        return self._storage_policy_name

    @property
    def copy_name(self):
        """returns copy name associated with store"""
        return self._copy_name

    @property
    def copy_id(self):
        """returns copy id the store is associated to"""
        return self._store_properties.get('copy').get('id')

    @property
    def enable_garbage_collection(self):
        """returns garbage collection property value for store"""
        if (self._extended_flags & 4) == 0:
            return False
        return True

    @property
    def enable_store_pruning(self):
        """returns if purning is enabled or disabled on store"""
        return self._store_flags & StoreFlags.IDX_SIDBSTORE_FLAGS_PRUNING_ENABLED.value != 0

    @enable_store_pruning.setter
    def enable_store_pruning(self, value):
        """sets store purning value to true or false
        Args:
              value (bool) -- value to enable or disable store pruning
        """
        if not value:
            new_value = self._store_flags & ~StoreFlags.IDX_SIDBSTORE_FLAGS_PRUNING_ENABLED.value
        else:
            new_value = self._store_flags | StoreFlags.IDX_SIDBSTORE_FLAGS_PRUNING_ENABLED.value

        request_json = {
            "EVGui_ParallelDedupConfigReq": {
                "processinginstructioninfo": "",
                "SIDBStore": {
                    "SIDBStoreId": self.store_id,
                    "SIDBStoreName": self.store_name,
                    "extendedFlags": self._extended_flags,
                    "flags": new_value,
                    "minObjSizeKB": 50,
                    "oldestEligibleObjArchiveTime": -1
                },
                "appTypeGroupId": 0,
                "commCellId": 2,
                "copyId": self.copy_id,
                "operation": 3
            }
        }
        output = self._commcell_object.qoperation_execute(request_json)
        if output['error']['errorString'] != '':
            raise SDKException('Storage', '102', output['error']['errorString'])

        self.refresh()

    @enable_garbage_collection.setter
    def enable_garbage_collection(self, value):
        """sets enable garbage collection with true or false
        Args:
              value (bool) -- value to enable or disable garbage collection
        """
        if self.version == -1:
            if not value:
                new_value = self._extended_flags & ~4
            else:
                new_value = self._extended_flags | 4

            request_json = {
                "EVGui_ParallelDedupConfigReq": {
                    "processinginstructioninfo": "",
                    "SIDBStore": {
                        "SIDBStoreId": self.store_id,
                        "SIDBStoreName": self.store_name,
                        "extendedFlags": new_value,
                        "flags": self._store_flags,
                        "minObjSizeKB": 50,
                        "oldestEligibleObjArchiveTime": -1
                    },
                    "appTypeGroupId": 0,
                    "commCellId": 2,
                    "copyId": self.copy_id,
                    "operation": 3
                }
            }
            self._commcell_object.qoperation_execute(request_json)
        self.refresh()

    def seal_deduplication_database(self):
        """ Seals the deduplication database """

        request_json = {
                        "App_SealSIDBStoreReq":{
                                "SidbStoreId": self.store_id
                            }
                        }
        self._commcell_object._qoperation_execute(request_json)

    def recover_deduplication_database(self, full_reconstruction=False, scalable_resources=True):
        """
        refresh store properties and start reconstruction job if at least one substore is marked for recovery

        Args:
            full_reconstruction (bool)  - to reconstruct without using previous backup (True/False)
                                        Default: False

            scalable_resources (bool)    - to run reconstruction using scalable resources
                                        Default: True

        Returns:
             object - instance of Job class for DDB Reconstruction job

        Raises:
             SDKException:
                if DDB Reconstruction job failed

                if response if empty

                if response in not success
        """
        self.refresh()
        substore_list = ""
        for substore in self.all_substores:
            if self._substores.get(substore[0]).get('status') == 1:
                substore_list += f"<SubStoreIdList val='{substore[0]}' />"

        if not substore_list:
            o_str = 'No substore is eligible for recon.'
            raise SDKException('Storage', '102', o_str)

        request_xml = f"""<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
                <TMMsg_CreateTaskReq>
                  <processinginstructioninfo/>
                  <taskInfo>
                    <task>
                      <taskFlags>
                        <disabled>false</disabled>
                      </taskFlags>
                      <policyType>DATA_PROTECTION</policyType>
                      <taskType>IMMEDIATE</taskType>
                      <initiatedFrom>COMMANDLINE</initiatedFrom>
                    </task>
                    <associations>
                      <copyName>{self.copy_name}</copyName>
                      <storagePolicyName>{self.storage_policy_name}</storagePolicyName>
                    </associations>
                    <subTasks>
                      <subTask>
                        <subTaskType>ADMIN</subTaskType>
                        <operationType>DEDUPDBSYNC</operationType>
                      </subTask>
                      <options>
                        <adminOpts>
                          <dedupDBSyncOption>
                            <SIDBStoreId>{self.store_id}</SIDBStoreId>
                            {substore_list}
                          </dedupDBSyncOption>
                          <reconstructDedupDBOption>
                            <noOfStreams>0</noOfStreams>
                            <allowMaximum>true</allowMaximum>
                            <flags>{int(full_reconstruction)}</flags>
                            <mediaAgents>
                              <mediaAgentName></mediaAgentName>
                            </mediaAgents>
                            <useScallableResourceManagement>{str(scalable_resources).lower()}</useScallableResourceManagement>
                          </reconstructDedupDBOption>
                        </adminOpts>
                      </options>
                      <subTaskOperation>OVERWRITE</subTaskOperation>
                    </subTasks>
                  </taskInfo>
                </TMMsg_CreateTaskReq>"""
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_xml
        )
        if flag:
            if response and response.json():
                if "jobIds" in response.json():
                    return Job(self._commcell_object, response.json()['jobIds'][0])
                elif "errorCode" in response.json():
                    error_message = response.json()['errorMessage']
                    o_str = 'DDB recon job failed\nError: "{0}"'.format(error_message)
                    raise SDKException('Storage', '102', o_str)
                raise SDKException('Storage', '112')
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def run_space_reclaimation(self, level=3, clean_orphan_data=False, use_scalable_resource=False):
        """
        runs DDB Space reclaimation job with provided level

        Args:
            level (int) - criteria for space reclaimation level (1, 2, 3, 4)
                        Default: 3

            clean_orphan_data (bool) - run space reclaimation with OCL or not (True/False)
                        Default: False

            use_scalable_resource (bool)    - Use Scalable Resource Allocation while running DDB Space Reclamation Job

        Returns:
             object - instance of Job class for DDB Verification job

        Raises:
             SDKException:
                if invalid level is provided

                if DDB space reclaimation job failed

                if response if empty

                if response in not success
        """
        if not (isinstance(level, int)) and level not in range(1, 4):
            raise SDKException('Storage', '101')

        if not isinstance(use_scalable_resource, bool):
            raise SDKException('Storage', '101')

        level_map = {
            1: 80,
            2: 60,
            3: 40,
            4: 20
        }
        clean_orphan_data = str(clean_orphan_data).lower()
        request_json = {
            "TMMsg_CreateTaskReq": {
                "taskInfo": {
                    "task": {
                        "taskFlags": {
                            "disabled": "false"
                        },
                        "policyType": "DATA_PROTECTION",
                        "taskType": "IMMEDIATE",
                        "initiatedFrom": "COMMANDLINE"
                    },
                    "associations": {
                        "copyName": self.copy_name,
                        "storagePolicyName": self.storage_policy_name,
                        "sidbStoreName": self.store_name
                    },
                    "subTasks": {
                        "subTask": {
                            "subTaskType": "ADMIN",
                            "operationType": "ARCHIVE_CHECK"
                        },
                        "options": {
                            "backupOpts": {
                                "mediaOpt": {
                                    "auxcopyJobOption": {
                                        "useMaximumStreams": "true",
                                        "maxNumberOfStreams": 0,
                                        "allCopies": "true",
                                        "mediaAgent": {
                                            "mediaAgentName": ""
                                        },
                                        "useScallableResourceManagement": f"{use_scalable_resource}"
                                    }
                                }
                            },
                            "adminOpts": {
                                "archiveCheckOption": {
                                    "ddbVerificationLevel": "DDB_DEFRAGMENTATION",
                                    "backupLevel": "FULL",
                                    "defragmentationPercentage": level_map.get(level),
                                    "ocl": clean_orphan_data
                                }
                            }
                        },
                        "subTaskOperation": "OVERWRITE"
                    }
                }
            }
        }
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self._commcell_object, response.json()['jobIds'][0])
                elif "errorCode" in response.json():
                    error_message = response.json()['errorMessage']

                    o_str = 'DDB space reclaimation job failed\nError: "{0}"'.format(error_message)
                    raise SDKException('Storage', '102', o_str)
                raise SDKException('Storage', '113')
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def run_ddb_verification(self, incremental_verification=True, quick_verification=True,
                             use_scalable_resource=False):
        """
        runs deduplication data verification(dv2) job with verification type and dv2 option

        Args:
            incremental_verification (bool) - DV2 job type, incremental or Full (True/False)
                                            Default: True (Incremental)

            quick_verification (bool)       - DV2 job option, Quick or Complete (True/False)
                                            Default: True (quick verification)

            use_scalable_resource (bool)    - Use Scalable Resource Allocation while running DDB Verification Job

        Returns:
             object - instance of Job class for DDB Verification job

        Raises:
             SDKException:
                if DDB Verification job failed

                if response if empty

                if response in not success
        """

        verification_type = 'INCREMENTAL'
        if not incremental_verification:
            verification_type = 'FULL'

        verification_option = 'QUICK_DDB_VERIFICATION'
        if not quick_verification:
            verification_option = 'DDB_AND_DATA_VERIFICATION'

        if not isinstance(use_scalable_resource, bool):
            raise SDKException('Storage', '101')

        request_json = {
            "TMMsg_CreateTaskReq": {
                "taskInfo": {
                    "task": {
                        "taskFlags": {
                            "disabled": "false"
                        },
                        "policyType": "DATA_PROTECTION",
                        "taskType": "IMMEDIATE",
                        "initiatedFrom": "COMMANDLINE"
                    },
                    "associations": {
                        "copyName": self.copy_name,
                        "storagePolicyName": self.storage_policy_name,
                        "sidbStoreName": self.store_name
                    },
                    "subTasks": {
                        "subTask": {
                            "subTaskType": "ADMIN",
                            "operationType": "ARCHIVE_CHECK"
                        },
                        "options": {
                            "backupOpts": {
                                "mediaOpt": {
                                    "auxcopyJobOption": {
                                        "useMaximumStreams": "true",
                                        "maxNumberOfStreams": 0,
                                        "allCopies": "true",
                                        "mediaAgent": {
                                            "mediaAgentName": ""
                                        },
                                        "useScallableResourceManagement": f"{use_scalable_resource}"
                                    }
                                }
                            },
                            "adminOpts": {
                                "archiveCheckOption": {
                                    "ddbVerificationLevel": verification_option,
                                    "backupLevel": verification_type
                                }
                            }
                        },
                        "subTaskOperation": "OVERWRITE"
                    }
                }
            }
        }
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._commcell_object._services['EXECUTE_QCOMMAND'], request_json
        )
        if flag:
            if response and response.json():
                if "jobIds" in response.json():
                    return Job(self._commcell_object, response.json()['jobIds'][0])
                elif "errorCode" in response.json():
                    error_message = response.json()['errorMessage']
                    o_str = 'DDB verification job failed\nError: "{0}"'.format(error_message)
                    raise SDKException('Storage', '102', o_str)
                raise SDKException('Storage', '108')
            raise SDKException('Response', '102')
        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)


class SubStore(object):
    """Class to performing substore level operations for Deduplication engine"""

    def __init__(self, commcell_object, storage_policy_name, copy_name, store_id, substore_id):
        """Initialise the SubStore class instance.

        Args:
            commcell_object (object)    - commcell class instance

            storage_policy_name (str)   - storage policy name in commcell

            copy_name (str)             - copy name under storage policy

            store_id (int)              - deduplication store id in commcell

            substore_id (int)           - substore id under deduplication store
        """
        self._commcell_object = commcell_object
        self._storage_policy_name = storage_policy_name
        self._copy_name = copy_name
        self._store_id = store_id
        self._substore_id = substore_id
        self._substore_properties = {}
        self._path = None
        self._media_agent = None
        self.refresh()

    def __repr__(self):
        """Representation string for the instance of the SubStore class."""
        return "SubStore class instance for Deduplication Substore ID: '{0}'".format(
            self._substore_id
        )

    def _initialize_substore_properties(self):
        """Initialize substore properties for the substore on a deduplication store"""
        store = Store(self._commcell_object, self._storage_policy_name, self._copy_name, self._store_id)
        self._substore_properties = store._substores[self._substore_id]
        self._path = self._substore_properties['Path']
        self._media_agent = self._substore_properties['MediaAgent']['name']

    def refresh(self):
        """refresh the properties of a substore"""
        self._initialize_substore_properties()

    def mark_for_recovery(self):
        """mark a substore for recovery and refresh substore properties"""
        request_json = {
            "EVGui_IdxSIDBSubStoreOpReq": {
                "info": {
                    "mediaAgent": {
                        "name": self.media_agent
                    },
                    "SIDBStoreId": self.store_id,
                    "SubStoreId": self.substore_id,
                    "opType": 1,
                    "path": self.path
                }
            }
        }
        self._commcell_object.qoperation_execute(request_json)
        self.refresh()

    @property
    def media_agent(self):
        """returns media agent for the substore"""
        return self._media_agent

    @property
    def path(self):
        """returns path for the substore"""
        return self._path

    @property
    def store_id(self):
        """returns store id for the substore"""
        return self._store_id

    @property
    def substore_id(self):
        """returns substore id"""
        return self._substore_id



"""
Main File  for Defining Logging class for CVC

"""
import os
import logging
from logging import Logger
from logging.handlers import RotatingFileHandler


class CVCLogger(object):
	"""
	Class for defining CVCLogger using Logging Module
	"""

	def __init__(self):
		self._log = None
		self._logfile = None

	@property
	def log(self):
		"""
		Returns Log attribute of CVCLogger object
		"""
		return self._log

	@log.setter
	def log(self, value):
		# Set the log
		self._log = value

	@property
	def logfile(self):
		"""
		Returns Logfile  attribute of CVCLogger object
		"""
		return self._logfile

	@logfile.setter
	def logfile(self, value):
		"""
		Sets Passed value to Attribute logfile of CVCLogger Object
		"""
		self._logfile = value

	def setuplogger(self, logfilename, loglocation):
		"""
		Method to set Location and Filename for respective log object
		"""
		return self.initializelogger(logfilename, loglocation)

	def initializelogger(self, logfilename, specifieddirectory):
		"""
		Method setting corresponding values for different attributes for Logging
		"""
		LOG_FORMAT = '%(process)d\t%(thread)d\t%(asctime)s\t%(levelname)-5s\t%(message)s'
		LOG_BYTE_SIZE = 5242880  # size in bytes of each log file.
		SAVED_LOG_COUNT = 5  # number of log files to keep
		LOG_WRITE_MODE = 'a'
		LOG_ENCODING = 'utf-8'
		creating_new_log_file = False
		try:
			if not os.path.exists(specifieddirectory):
				os.makedirs(specifieddirectory)

			logFile = os.path.join(specifieddirectory, logfilename)
			if not os.path.exists(logFile):
				creating_new_log_file = True

			try:
				# Initialize the logger
				logger = Logger(logfilename.split('.')[0])
				loglevel = logging.DEBUG
				filehandler = RotatingFileHandler(
					logFile,
					LOG_WRITE_MODE,
					LOG_BYTE_SIZE,
					SAVED_LOG_COUNT,
					encoding=LOG_ENCODING)
				filehandler.setLevel(loglevel)
				filehandler.setFormatter(logging.Formatter(LOG_FORMAT))
				logger.addHandler(filehandler)
				# once the log file is created successfully, we set the permisssions to 666
				if os.path.exists(logFile) and creating_new_log_file:
					import platform
					if platform.system() == 'Linux' or platform.system() == 'OS400':
						os.chmod(logFile, 0o666)
			except Exception as erno:
				print(erno)
				raise

			# Set the loggers.
			global log
			self.log = logger  # This allows creator access to the object
			self.logfile = logFile
			log = logger  # This allows future importers access to the object

			return True
		except Exception as err:
			print(err)
			raise



"""Main file for performing Local Simpana Detection Operation """

import os
import re
import platform
import subprocess
import logger
try:
    import winreg
except ImportError:
    pass


class Simpanadefaults():
    """
    Class Simpana Defaults is responsible for Local Commvault detection
    Methods :
    local_cs_detection : Method to detect local Commvault  installation and fetch CS name
    local_client_detection : Method to detect local Commvault  installation and fetch Clientname
    """
    @staticmethod
    def _winregistry_reading(path, keyname):
        """
        Method to read Windows Registry
        """
        key_value = None
        status = False
        try:
            key2 = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                  path, 0, winreg.KEY_ALL_ACCESS)
            key_list = winreg.QueryValueEx(key2, keyname)
            key_value = key_list[0]
            status = True
        except Exception:
            status = False
        finally:
            return status, key_value

    def local_cs_detection(self, instance_name):
        """
        Method to detect local commvault  installation and
        returns the local  CS name  to the calling function
        Args :
        instance_name:Commvault Instance
        Returns :
        status = True | False depending on Commvault Detection on Local Machine
        csName= Commvault CS for the currently detected Local client
        """
        loglines = []
        try:
            csname = ""
            flag = False
            keyvalue2 = ""
            INSTANCE_PATH = 'SOFTWARE\\CommVault Systems\\Galaxy\\' + instance_name
            if platform.system() == 'Windows':
                comm_path = INSTANCE_PATH + '\\CommServe'
                keyname = 'sCSHOSTNAME'
                (status, cs_hostname) = Simpanadefaults(
                )._winregistry_reading(comm_path, keyname)
                if status is True:
                    loglines.append(["i", "CS Hostname is :%s" % cs_hostname])
                    csname = cs_hostname
                    flag = True
                else:
                    loglines.append(["e", "Issue in reading from Windows registry"])
            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    linux_path1 = reg_path + instance_name
                    if os.path.exists(linux_path1):
                        linux_path2 = reg_path + instance_name + "/CommServe"
                        if os.path.exists(linux_path2):
                            filename2 = linux_path2 + "/.properties"
                            file2 = open(filename2, 'r')
                            lines = file2.readlines()
                            for item in lines:
                                item = str(item)
                                if item.startswith("sCSHOSTNAME"):
                                    reg_detail2 = item
                                    break
                                else:
                                    continue
                            cs_node_name = reg_detail2.split()
                            keyvalue2 = str(cs_node_name[1])
                            csname = keyvalue2
                            file2.close()
                            flag = True
                        else:
                            loglines.append(["e",
                                             "Commvault Registry Path does not exist : " + str(linux_path2)])
                            # sys.exit(0)
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : " + str(linux_path1)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Error in Opening Linux Commvault Registry"])
                    loglines.append(["e", "Exception is %s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                loglines.append(["i", 'IBMi machine, skipping...'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in local CS detection : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, csname, loglines)

    def local_client_detection(self, instance_name):
        """
        Method to detect local commvault installation  and
        returns the local client name to the calling function
        Args :
        instance_name:Commvault Instance
        Returns :
        status = True | False depending on Commvault Detection on Local Machine
        clientName=Local client name
        """
        loglines = []
        try:
            clientname = ""
            flag = False
            INSTANCE_PATH = 'SOFTWARE\\CommVault Systems\\Galaxy\\' + instance_name
            if platform.system() == 'Windows':
                keyname = 'sPhysicalNodeName'
                (status, clientname) = Simpanadefaults(
                )._winregistry_reading(INSTANCE_PATH, keyname)
                if status is True:
                    flag = True
                else:
                    loglines.append(["e", "Issue in reading from Windows registry"])

            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    linux_path1 = reg_path + instance_name
                    if os.path.exists(linux_path1):
                        filename1 = linux_path1 + "/.properties"
                        file1 = open(filename1, 'r')
                        lines = file1.readlines()
                        for item in lines:
                            item = str(item)
                            if item.startswith("sPhysicalNodeName"):
                                reg_detail1 = item
                                break
                            else:
                                continue
                        physical_node_name = reg_detail1.split()
                        keyvalue1 = str(physical_node_name[1])
                        clientname = keyvalue1
                        flag = True
                        file1.close()
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : %s" % (linux_path1)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Exception is :%s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                loglines.append(["i", 'IBMi machine, skipping...'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in client detection : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, clientname, loglines)

    def check_for_local_instance(self, client_name):
        """
        Method to detect if client passed belongs to local instance
        Args :
        client_name:Client Name
        Returns :
        status = True | False depending on whether client is local
        """
        loglines = []
        try:
            flag = False

            if platform.system() == 'Windows':
                instances_list = []
                win_path = 'SOFTWARE\\CommVault Systems\\Galaxy'
                try:
                    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                         win_path, 0, winreg.KEY_ALL_ACCESS)
                    i = 0
                    while True:
                        try:
                            subkey = winreg.EnumKey(key, i)
                            instances_list.append(subkey)
                            i = i + 1
                        except WindowsError:
                            break
                    for path in instances_list:
                        if path.startswith("Instance"):
                            instance_path = 'SOFTWARE\\CommVault Systems\\Galaxy' + "\\" + path
                            keyname = 'sPhysicalNodeName'
                            (status, clientname) = Simpanadefaults(
                            )._winregistry_reading(instance_path, keyname)
                            if status is True:
                                if clientname == client_name:
                                    flag = True
                                    break
                except Exception as windows_registry_exception:
                    windows_registry_exception = str(
                        windows_registry_exception)
                    flag = False
                    loglines.append(["e", "Exception in accessing Windows Commvault Registry :%s" %
                                     windows_registry_exception])

            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    if os.path.exists(reg_path):
                        instances_list = None
                        instances_list = os.listdir(reg_path)
                        if instances_list:
                            for i in instances_list:
                                if i.startswith("Instance"):
                                    filename1 = reg_path + i + "/.properties"
                                    file1 = open(filename1, 'r')
                                    lines = file1.readlines()
                                    for item in lines:
                                        item = str(item)
                                        if item.startswith("sPhysicalNodeName"):
                                            reg_detail1 = item
                                            break
                                        else:
                                            continue
                                    physical_node_name = reg_detail1.split()
                                    keyvalue1 = str(physical_node_name[1])
                                    clientname = keyvalue1
                                    if clientname == client_name:
                                        flag = True
                                        file1.close()
                                        break
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : %s" % (reg_path)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Exception is :%s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                loglines.append(["i", 'IBMi machine, skipping...'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in checking for local instance : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, loglines)

    def simpana_installation_path(self, instance_name):
        loglines = []

        try:
            installation_path = ""
            flag = False
            keyvalue1 = ""
            if platform.system() == 'Windows':
                base_path = 'SOFTWARE\\CommVault Systems\\Galaxy' + \
                    '\\' + instance_name + '\\' + 'Base'
                base_path = str(base_path)
                try:
                    key1 = winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE, base_path, 0, winreg.KEY_READ)
                    galaxy_home = winreg.QueryValueEx(
                        key1, "dGALAXYHOME")
                    keyvalue1 = galaxy_home[0]
                    installation_path = keyvalue1
                    flag = True
                except Exception as windows_registry_exception:
                    windows_registry_exception = str(
                        windows_registry_exception)
                    flag = False
                    loglines.append(["e", "Exception in accessing Windows Commvault Registry :%s" %
                                     windows_registry_exception])

            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    linux_path1 = reg_path + instance_name + "/Base"
                    if os.path.exists(linux_path1):
                        filename1 = linux_path1 + "/.properties"
                        file1 = open(filename1, 'r')
                        lines = file1.readlines()
                        for item in lines:
                            item = str(item)
                            if item.startswith("dGALAXYHOME"):
                                reg_detail1 = item
                                break
                            else:
                                continue
                        galaxy_home = reg_detail1.split()
                        keyvalue1 = str(galaxy_home[1])
                        installation_path = keyvalue1
                        file1.close()
                        flag = True
                        # sys.exit(0)
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : %s" % (linux_path1)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Exception is :%s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                loglines.append(["i", 'IBMi machine, skipping...'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in Commvault installation path detection : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, installation_path, loglines)

    def simpana_logging_path(self, instance_name):
        loglines = []

        try:
            logging_path = ""
            flag = False
            keyvalue1 = ""
            if platform.system() == 'Windows':
                base_path = 'SOFTWARE\\CommVault Systems\\Galaxy' + '\\' + instance_name + '\\' + 'EventManager'
                base_path = str(base_path)
                try:
                    key1 = winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE, base_path, 0, winreg.KEY_READ)
                    galaxy_home = winreg.QueryValueEx(
                        key1, "dEVLOGDIR")
                    keyvalue1 = galaxy_home[0]
                    logging_path = keyvalue1
                    flag = True
                except Exception as windows_registry_exception:
                    windows_registry_exception = str(
                        windows_registry_exception)
                    flag = False
                    loglines.append(["e", "Exception in accessing Windows Commvault Registry :%s" %
                                     windows_registry_exception])

            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    linux_path1 = reg_path + instance_name + "/EventManager"
                    if os.path.exists(linux_path1):
                        filename1 = linux_path1 + "/.properties"
                        file1 = open(filename1, 'r')
                        lines = file1.readlines()
                        for item in lines:
                            item = str(item)
                            if item.startswith("dEVLOGDIR"):
                                reg_detail1 = item
                                break
                            else:
                                continue
                        galaxy_home = reg_detail1.split()
                        keyvalue1 = str(galaxy_home[1])
                        logging_path = keyvalue1
                        file1.close()
                        flag = True
                        # sys.exit(0)
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : %s" % (linux_path1)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Exception is :%s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                loglines.append(["i", 'IBMi machine, skipping...'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in Commvault logging path detection : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, logging_path, loglines)

    def check_valid_instance(self, instance_name):
        """
        Method to detect if instance passed is a valid instance
        Args :
        instance_name:instance Name
        Returns :
        status = True | False depending on whether instance is valid
        """
        loglines = []
        loglines.append(["i", "Checking validity of instance name"])
        try:
            flag = False
            if platform.system() == 'Windows':
                instances_list = []
                win_path = 'SOFTWARE\\CommVault Systems\\Galaxy'
                try:
                    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                         win_path, 0, winreg.KEY_ALL_ACCESS)
                    i = 0
                    while True:
                        try:
                            subkey = winreg.EnumKey(key, i)
                            instances_list.append(subkey)
                            i = i + 1
                        except WindowsError:
                            break
                    for p in instances_list:
                        if p.startswith("Instance"):
                            if p == instance_name:
                                flag = True
                                loglines.append(["i", "Instance name found to be valid"])
                                break
                except Exception as windows_registry_exception:
                    windows_registry_exception = str(
                        windows_registry_exception)
                    flag = False
                    loglines.append(["e", "Exception in accessing Windows Commvault Registry :%s" %
                                     windows_registry_exception])

            elif platform.system() == 'Linux':
                try:
                    cmd = subprocess.Popen(
                        "id -u", stdout=subprocess.PIPE, shell=True)
                    i, error = cmd.communicate()
                    i = int(i)
                    reg_path = "/etc/CommVaultRegistry/Galaxy/"
                    if i != 0:
                        user_cmd = subprocess.Popen(
                            "id -u -n", stdout=subprocess.PIPE, shell=True)
                        user, error = user_cmd.communicate()
                        home_cmd = subprocess.Popen(
                            "echo $HOME", stdout=subprocess.PIPE, shell=True)
                        home, error = home_cmd.communicate()
                        home_string = home.decode().rstrip('\n')
                        cv_reg_path = "CommVaultRegistry/Galaxy/"
                        cv_reg_path = os.path.join(home_string, cv_reg_path)
                        if os.path.exists(cv_reg_path):
                            reg_path = cv_reg_path
                    if os.path.exists(reg_path):
                        instances_list = None
                        instances_list = os.listdir(reg_path)
                        if instances_list:
                            for i in instances_list:
                                if i.startswith("Instance"):
                                    if i == instance_name:
                                        flag = True
                                        loglines.append(["i", "Instance name found to be valid"])
                                        break
                    else:
                        loglines.append(["e",
                                         "Commvault Registry Path does not exist : %s" % (reg_path)])

                except (IOError, SystemError) as linux_registry_exception:
                    loglines.append(["e", "Exception is :%s" % linux_registry_exception])
                    flag = False
            elif platform.system() == 'OS400':
                flag = True
                loglines.append(["i", 'IBMi machine, skipping instance validation'])
            else:
                loglines.append(["e", "Unsupported OS"])
                # sys.exit(0)
        except Exception as simpana_detect_exception:
            loglines.append(["e", "Exception in checking for valid instance : %s" %
                             simpana_detect_exception])

        finally:
            return (flag, loglines)

    def get_local_instance_name(self, install_directory):
        """To get the local instance name based on install directory provided

        Args:
            install_directory	(str)	-- Commvault install directory

        Returns:
            (str)	-- Instance name

        """
        os_type = platform.system()
        if os_type == 'Linux':
            with open('{0}/galaxy_vm'.format(install_directory), 'r') as file:
                content = file.read()
            temp = re.findall('GALAXY_INST="(.+?)";', content)
            if temp:
                return temp[0]
        elif os_type == 'Windows':
            with open(r'{0}\Base\QinetixVM'.format(install_directory), 'r') as file:
                content = file.read()
            if content:
                return content.strip()

        return "Instance001"

    def allow_http(self, instance, log):
        """
        Checks for registry key value of bAllowCVCHttpFallback

        Args:

            instance    (str)   --  Instance name like Instance001

            log         (obj)   --  Logger class object

        Returns:

            (bool)  -   True if bAllowCVCHttpFallback is set True else false

        """
        log.info("Checking registry value for key bAllowCVCHttpFallback")
        value = None
        try:
            if platform.system() == 'Windows':
                reg_path = rf'SOFTWARE\CommVault Systems\Galaxy\{instance}\EventManager'
                _, value = self._winregistry_reading(reg_path, 'bAllowCVCHttpFallback')

            elif platform.system() == 'Linux':
                value = self._get_unix_registry(instance, 'EventManager', 'bAllowCVCHttpFallback')

        except Exception:
            log.info("Unable to fetch registry value for bAllowCVCHttpFallback")

        finally:
            if value:
                return True
            return False

    def _get_unix_registry(self, instance, category, key_name):
        """
        Reads the registry value from the given path

        Args:
            instance    (str)   --  Instance name like Instance001

            category    (str)   -- Category the key belongs to
                                    Example: EventManager

            key_name    (str)   -- Registry key name

        """
        value = None
        cmd = subprocess.Popen("id -u", stdout=subprocess.PIPE, shell=True)
        user_id, _ = cmd.communicate()
        user_id = int(user_id)
        reg_path = "/etc/CommVaultRegistry/Galaxy"
        if user_id != 0:
            home_cmd = subprocess.Popen("echo $HOME", stdout=subprocess.PIPE, shell=True)
            home_path, _ = home_cmd.communicate()
            home_path = home_path.decode().rstrip('\n')
            temp_reg_path = f"{home_path}/CommVaultRegistry/Galaxy"

            if os.path.exists(temp_reg_path):
                reg_path = temp_reg_path

        reg_path = f"{reg_path}/{instance}/{category}/.properties"

        if not os.access(reg_path, 4):
            return ''
        if os.path.exists(reg_path):
            with open(reg_path, 'r') as rfile:
                content = rfile.readlines()

            for line in content:
                if line.startswith(key_name):
                    value = line.split(' ')[1]
                    try:
                        value = int(value)
                    except Exception:
                        pass

                    break

        return value



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Cloud Apps Instance.

CloudAppsInstance is the only class defined in this file.

CloudAppsInstance:  Derived class from Instance Base class, representing a
cloud apps instance, and to perform operations on that instance

**Note:** GoogleInstance class is used for OneDrive as well.

CloudAppsInstance:

    __new__()   --  Method to create object based on specific cloud apps instance type


Usage
=====

To add a new Instance for Cloud Apps agent, please follow these steps:

    1. Add the module for the new instance type under the location:
        **/cvpysdk/instances/cloudapps**,
        with the module name **<new instance type>_instance.py**
        (e.g. "google_instance.py", "salesforce_instance.py")

    #. Create a class for your instance type and inherit the CloudAppsInstance class.

    #. Add the import statement inside the __new__ method.
        **NOTE:** If you add the import statement at the top,
        it'll cause cyclic import, and the call will fail

    #. After adding the import statement:
        - In the **instance_type** dict
            - Add the cloud apps instance type as the key, and the class as its value

"""

from __future__ import unicode_literals

from ..instance import Instance
from ..exception import SDKException


class CloudAppsInstance(Instance):
    """Class for representing an Instance of the Cloud Apps agent."""

    def __new__(cls, agent_object, instance_name, instance_id):
        from .cloudapps.google_instance import GoogleInstance
        from .cloudapps.salesforce_instance import SalesforceInstance
        from .cloudapps.cloud_storage_instance import CloudStorageInstance
        from .cloudapps.amazon_instance import AmazonRedshiftInstance
        from .cloudapps.amazon_instance import AmazonDocumentDBInstance
        from .cloudapps.amazon_instance import AmazonRDSInstance
        from .cloudapps.amazon_instance import AmazonDynamoDBInstance
        from .cloudapps.dynamics365_instance import MSDynamics365Instance
        from .cloudapps.teams_instance import TeamsInstance
        from .cloudapps.spanner_instance import GoogleSpannerInstance
        instance_type = {
            1: GoogleInstance,
            2: GoogleInstance,
            3: SalesforceInstance,
            4: AmazonRDSInstance,     # Amazon RDS Instance
            5: CloudStorageInstance,  # AmazonS3 Instance
            6: CloudStorageInstance,  # AzureBlob Instance
            # OneDrive Instance, GoogleInstance class is used for OneDrive instance too.
            7: GoogleInstance,
            14: CloudStorageInstance,  # OracleCloud Instance
            15: CloudStorageInstance,  # Openstack Instance
            20: CloudStorageInstance,  # Google Cloud Instance
            21: CloudStorageInstance,  # azure data lake gen2
            26: AmazonRedshiftInstance, # Amazon Redshift
            27: AmazonDocumentDBInstance, # Amazon Document DB
            25: CloudStorageInstance,   #AliBaba
            24: CloudStorageInstance,   #IBM
            22: AmazonDynamoDBInstance, # Amazon DynamoDB
            35: MSDynamics365Instance,   #  MS Dynamics 365 Instance
            36: TeamsInstance, # Office 365 Teams
			37: GoogleSpannerInstance # Google Cloud Spanner Instance
        }

        commcell_object = agent_object._commcell_object
        instance_service = 'Instance/{0}'.format(instance_id)

        response = commcell_object.request('GET', instance_service)

        if response.json() and "instanceProperties" in response.json():
            properties = response.json()["instanceProperties"][0]
        else:
            raise SDKException('Instance', '105')

        cloud_apps_instance_type = properties['cloudAppsInstance']['instanceType']

        return object.__new__(instance_type[cloud_apps_instance_type])



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing system related operations on Commcell.

System is the only class defined in this file

System: Class for performing system related operations on the commcell

System:
=======

    __init__(commcell_object)       --  initialise object of System class

    set_gui_timeout()               --  To set GUI timeout value in minutes

"""


class System:
    """Class for performing system related operations in the commcell"""

    def __init__(self, commcell_object):
        """Initialize the System class

            Args:
                commcell_object    (object)    --  instance of the Commcell class

        """
        self._commcell_object = commcell_object

    def set_gui_timeout(self, value):
        """Sets GUI timeout value in minutes

        Args:
            value   (str)   -- GUI timeout value in minutes

        **Note** setting value to 0 will disable GUI connections to timeout

        """
        self._commcell_object._set_gxglobalparam_value(
            {
                "name": "Gui timeout",
                "value": str(value)
            }
        )
